# mautrix-twitter - A Matrix-Twitter DM puppeting bridge
# Copyright (C) 2022 Tulir Asokan, Max Sandholm
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
from __future__ import annotations

import asyncio
import logging
import traceback

# from .config import Config
from .db import BackfillStatus as DBBackfillStatus
from .portal import Portal
from .user import User


class NoFirstMessageException(BaseException):
    pass


class BackfillStatus(DBBackfillStatus):
    recheck_queues: set[asyncio.Queue] = set()

    @classmethod
    async def recheck(cls) -> None:
        for q in cls.recheck_queues:
            await q.put(True)

    @classmethod
    async def get_next_backfill_status(cls, recheck_queue: asyncio.Queue) -> BackfillStatus:
        while True:
            status = await cls.get_next_unfinished_status()
            if status != None:
                return status

            try:
                await asyncio.wait_for(recheck_queue.get(), 10)
            except asyncio.exceptions.TimeoutError:
                pass

    @classmethod
    async def backfill_loop(cls) -> None:
        recheck_queue = asyncio.Queue()
        cls.recheck_queues.add(recheck_queue)

        while True:
            await asyncio.sleep(2)
            state = await cls.get_next_backfill_status(recheck_queue)
            portal = await Portal.get_by_twid(twid=state.twid, receiver=state.receiver)
            source = await User.get_by_twid(state.backfill_user)
            try:
                state.dispatched = True
                await state.update()
                num_filled = await portal.backfill(source, is_initial=state.state == 0)

                state.message_count += num_filled
                if num_filled == 0:
                    state.state = 2
                elif state.state == 0:
                    state.state = 1
            except NoFirstMessageException:
                logging.error("No first message found to do backfill!")
                state.state = 0
            except Exception:
                # TODO: handle and log error, and be smarter about backfill loops
                state.state = 3
                logging.error(traceback.format_exc())
            finally:
                state.dispatched = False
                await state.update()



# mautrix-twitter - A Matrix-Twitter DM puppeting bridge
# Copyright (C) 2022 Tulir Asokan, Max Sandholm
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
from __future__ import annotations

from typing import TYPE_CHECKING, ClassVar

from attr import dataclass

from mautrix.util.async_db import Database

fake_db = Database.create("") if TYPE_CHECKING else None


@dataclass
class BackfillStatus:
    db: ClassVar[Database] = fake_db

    twid: str
    receiver: int | None
    backfill_user: int
    dispatched: bool
    message_count: int
    state: int

    @property
    def _values(self):
        return (
            self.twid,
            self.receiver,
            self.backfill_user,
            self.dispatched,
            self.message_count,
            self.state,
        )

    async def insert(self) -> None:
        q = """INSERT INTO backfill_status (twid, receiver, backfill_user, dispatched, message_count, state)
            VALUES ($1, $2, $3, $4, $5, $6)"""
        await self.db.execute(q, *self._values)

    async def update(self) -> None:
        q = """UPDATE backfill_status SET backfill_user=$3, dispatched=$4, message_count=$5, state=$6
            WHERE twid=$1 AND receiver=$2"""
        await self.db.execute(q, *self._values)

    @classmethod
    async def get_by_twid(cls, twid: int, receiver: int = 0) -> BackfillStatus | None:
        q = """SELECT twid, receiver, backfill_user, dispatched, message_count, state FROM backfill_status
        WHERE twid =$1 AND receiver =$2"""

        row = await cls.db.fetchrow(q, twid, receiver)
        if not row:
            return None
        return cls(**row)

    @classmethod
    async def get_next_unfinished_status(cls) -> BackfillStatus | None:
        q = """SELECT twid, receiver, backfill_user, dispatched, message_count, state FROM backfill_status
        WHERE dispatched IS FALSE
        AND state < 2
        ORDER BY state ASC, message_count ASC
        LIMIT 1"""
        row = await cls.db.fetchrow(q)
        if not row:
            return None
        return cls(**row)



# mautrix-twitter - A Matrix-Twitter DM puppeting bridge
# Copyright (C) 2020 Tulir Asokan
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
from asyncpg import Connection

from mautrix.util.async_db import Scheme, UpgradeTable

upgrade_table = UpgradeTable()


@upgrade_table.register(description="Latest revision", upgrades_to=8)
async def upgrade_latest(conn: Connection, scheme: Scheme) -> None:
    if scheme in (Scheme.POSTGRES, Scheme.COCKROACH):
        await conn.execute("CREATE TYPE twitter_conv_type AS ENUM ('ONE_TO_ONE', 'GROUP_DM')")
    await conn.execute(
        """CREATE TABLE portal (
            twid        TEXT,
            receiver    BIGINT,
            conv_type   twitter_conv_type NOT NULL,
            other_user  BIGINT,
            mxid        TEXT,
            name        TEXT,
            encrypted   BOOLEAN NOT NULL DEFAULT false,

            next_batch_id TEXT,

            PRIMARY KEY (twid, receiver)
        )"""
    )
    await conn.execute(
        """CREATE TABLE "user" (
            mxid        TEXT PRIMARY KEY,
            twid        BIGINT,
            auth_token  TEXT,
            csrf_token  TEXT,
            poll_cursor TEXT,
            notice_room TEXT
        )"""
    )
    await conn.execute(
        """CREATE TABLE puppet (
            twid      BIGINT PRIMARY KEY,
            name      TEXT,
            photo_url TEXT,
            photo_mxc TEXT,

            is_registered BOOLEAN NOT NULL DEFAULT false,
            contact_info_set BOOLEAN NOT NULL DEFAULT false,

            custom_mxid  TEXT,
            access_token TEXT,
            next_batch   TEXT,
            base_url     TEXT
        )"""
    )
    await conn.execute(
        """CREATE TABLE message (
            mxid     TEXT NOT NULL,
            mx_room  TEXT NOT NULL,
            twid     BIGINT,
            receiver BIGINT,

            PRIMARY KEY (twid, receiver),
            UNIQUE (mxid, mx_room)
        )"""
    )
    await conn.execute(
        """CREATE TABLE reaction (
            mxid        TEXT NOT NULL,
            mx_room     TEXT NOT NULL,
            tw_msgid    BIGINT,
            tw_receiver BIGINT,
            tw_sender   BIGINT,
            reaction    TEXT NOT NULL,

            tw_reaction_id BIGINT,

            PRIMARY KEY (tw_msgid, tw_receiver, tw_sender),
            FOREIGN KEY (tw_msgid, tw_receiver) REFERENCES message(twid, receiver)
                ON DELETE CASCADE ON UPDATE CASCADE,
            UNIQUE (mxid, mx_room)
        )"""
    )
    await conn.execute(
        """CREATE TABLE backfill_status (
             twid TEXT,
             receiver BIGINT,
             backfill_user BIGINT,
             dispatched BOOLEAN,
             message_count INTEGER,
             state INTEGER,
             PRIMARY KEY (twid, receiver),
             FOREIGN KEY (twid, receiver) REFERENCES portal(twid, receiver)
                 ON DELETE CASCADE
        )"""
    )


@upgrade_table.register(description="Add double-puppeting base_url to puppet table")
async def upgrade_v2(conn: Connection) -> None:
    await conn.execute("ALTER TABLE puppet ADD COLUMN base_url TEXT")


@upgrade_table.register(description="Store Twitter reaction IDs for marking things read")
async def upgrade_v3(conn: Connection) -> None:
    await conn.execute("ALTER TABLE reaction ADD COLUMN tw_reaction_id BIGINT")


@upgrade_table.register(description="Replace VARCHAR(255) with TEXT")
async def upgrade_v4(conn: Connection) -> None:
    tables = {
        "portal": ("twid", "mxid", "name"),
        "user": ("mxid", "auth_token", "csrf_token", "poll_cursor", "notice_room"),
        "puppet": ("name", "photo_url", "photo_mxc", "custom_mxid", "next_batch"),
        "message": ("mxid", "mx_room"),
        "reaction": ("mxid", "mx_room"),
    }
    for table, columns in tables.items():
        for column in columns:
            await conn.execute(f'ALTER TABLE "{table}" ALTER COLUMN "{column}" TYPE TEXT')
    await conn.execute("DROP TABLE user_portal")


@upgrade_table.register(description="Add table for backfill status")
async def upgrade_v5(conn: Connection) -> None:
    await conn.execute("""ALTER TABLE portal ADD COLUMN next_batch_id TEXT""")

    await conn.execute(
        """CREATE TABLE backfill_status (
             twid TEXT,
             receiver BIGINT,
             backfill_user BIGINT,
             dispatched BOOLEAN,
             message_count INTEGER,
             state INTEGER,
             PRIMARY KEY (twid, receiver),
             FOREIGN KEY (twid, receiver) REFERENCES portal(twid, receiver)
                 ON DELETE CASCADE
        )"""
    )

    # For any existing portals, don't backfill.
    # This inserts rows that say those portals are already completed.
    await conn.execute(
        """
        INSERT INTO backfill_status (twid, receiver, backfill_user, dispatched, message_count, state)
        SELECT twid, receiver, 0, FALSE, 0, 3 FROM portal
        """
    )


@upgrade_table.register(description="Add contact_info_set column to puppet table")
async def upgrade_v6(conn: Connection) -> None:
    await conn.execute(
        "ALTER TABLE puppet ADD COLUMN contact_info_set BOOLEAN NOT NULL DEFAULT false"
    )


@upgrade_table.register(description="Reset the poll_cursor to force contact info sync")
async def upgrade_v7(conn: Connection) -> None:
    await conn.execute('UPDATE "user" SET poll_cursor = NULL')


@upgrade_table.register(description="Remove emoji type enum")
async def upgrade_v8(conn: Connection, scheme: Scheme) -> None:
    if scheme in (Scheme.COCKROACH, Scheme.POSTGRES):
        await conn.execute("ALTER TABLE reaction ALTER COLUMN reaction TYPE TEXT")
        await conn.execute("DROP TYPE twitter_reaction_key")
    await conn.execute(
        """UPDATE reaction SET reaction=CASE
            WHEN reaction='funny' THEN '😂'
            WHEN reaction='surprised' THEN '😲'
            WHEN reaction='sad' THEN '😢'
            WHEN reaction='like' THEN '❤'
            WHEN reaction='excited' THEN '🔥'
            WHEN reaction='agree' THEN '👍'
            WHEN reaction='disagree' THEN '👎'
            WHEN reaction='emoji' THEN ''
        END"""
    )



# mautrix-twitter - A Matrix-Twitter DM puppeting bridge
# Copyright (C) 2022 Tulir Asokan
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
from __future__ import annotations

from typing import TYPE_CHECKING, ClassVar

from attr import dataclass
import asyncpg

from mautrix.types import EventID, RoomID
from mautrix.util.async_db import Database
from mautwitdm.types import ReactionKey

fake_db = Database.create("") if TYPE_CHECKING else None


@dataclass
class Reaction:
    db: ClassVar[Database] = fake_db

    mxid: EventID
    mx_room: RoomID
    tw_msgid: int
    tw_receiver: int
    tw_sender: int
    reaction: str
    tw_reaction_id: int

    async def insert(self) -> None:
        q = (
            "INSERT INTO reaction (mxid, mx_room, tw_msgid, tw_receiver, tw_sender, reaction, tw_reaction_id) "
            "VALUES ($1, $2, $3, $4, $5, $6, $7)"
        )
        await self.db.execute(
            q,
            self.mxid,
            self.mx_room,
            self.tw_msgid,
            self.tw_receiver,
            self.tw_sender,
            self.reaction,
            self.tw_reaction_id,
        )

    async def update_id(self, tw_reaction_id: int) -> None:
        q = (
            "UPDATE reaction SET tw_reaction_id=$1 "
            "WHERE tw_msgid=$4 AND tw_receiver=$5 AND tw_sender=$6"
        )
        await self.db.execute(q, tw_reaction_id, self.tw_msgid, self.tw_receiver, self.tw_sender)

    async def edit(
        self, mx_room: RoomID, mxid: EventID, reaction: str, tw_reaction_id: int | None
    ) -> None:
        q = (
            "UPDATE reaction SET mxid=$1, mx_room=$2, reaction=$3, tw_reaction_id=$7 "
            "WHERE tw_msgid=$4 AND tw_receiver=$5 AND tw_sender=$6"
        )
        await self.db.execute(
            q,
            mxid,
            mx_room,
            reaction,
            self.tw_msgid,
            self.tw_receiver,
            self.tw_sender,
            tw_reaction_id,
        )

    async def delete(self) -> None:
        q = "DELETE FROM reaction WHERE tw_msgid=$1 AND tw_receiver=$2 AND tw_sender=$3"
        await self.db.execute(q, self.tw_msgid, self.tw_receiver, self.tw_sender)

    @classmethod
    def _from_row(cls, row: asyncpg.Record) -> Reaction | None:
        if not row:
            return None
        return cls(**row)

    @classmethod
    async def get_by_mxid(cls, mxid: EventID, mx_room: RoomID) -> Reaction | None:
        q = (
            "SELECT mxid, mx_room, tw_msgid, tw_receiver, tw_sender, reaction, tw_reaction_id "
            "FROM reaction WHERE mxid=$1 AND mx_room=$2"
        )
        return cls._from_row(await cls.db.fetchrow(q, mxid, mx_room))

    @classmethod
    async def get_last(cls, mx_room: RoomID) -> Reaction | None:
        q = (
            "SELECT mxid, mx_room, tw_msgid, tw_receiver, tw_sender, reaction, tw_reaction_id "
            "FROM reaction WHERE mx_room=$1 and tw_reaction_id IS NOT NULL ORDER BY tw_reaction_id DESC LIMIT 1"
        )
        return cls._from_row(await cls.db.fetchrow(q, mx_room))

    @classmethod
    async def get_by_message_twid(
        cls,
        tw_msgid: int,
        tw_receiver: int,
        tw_sender: int,
    ) -> Reaction | None:
        q = (
            "SELECT mxid, mx_room, tw_msgid, tw_receiver, tw_sender, reaction, tw_reaction_id "
            "FROM reaction WHERE tw_msgid=$1 AND tw_sender=$2 AND tw_receiver=$3"
        )
        return cls._from_row(await cls.db.fetchrow(q, tw_msgid, tw_sender, tw_receiver))

    @classmethod
    async def get_by_reaction_twid(cls, tw_reaction_id: int, tw_receiver: int) -> Reaction | None:
        q = (
            "SELECT mxid, mx_room, tw_msgid, tw_receiver, tw_sender, reaction, tw_reaction_id "
            "FROM reaction WHERE tw_reaction_id=$1 AND tw_receiver=$2"
        )
        return cls._from_row(await cls.db.fetchrow(q, tw_reaction_id, tw_receiver))



# mautrix-twitter - A Matrix-Twitter DM puppeting bridge
# Copyright (C) 2022 Tulir Asokan
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
from __future__ import annotations

from typing import TYPE_CHECKING, ClassVar

from attr import dataclass

from mautrix.types import RoomID, UserID
from mautrix.util.async_db import Database

fake_db = Database.create("") if TYPE_CHECKING else None


@dataclass
class User:
    db: ClassVar[Database] = fake_db

    mxid: UserID
    twid: int | None
    auth_token: str | None
    csrf_token: str | None
    poll_cursor: str | None
    notice_room: RoomID | None

    @property
    def _values(self):
        return (
            self.mxid,
            self.twid,
            self.auth_token,
            self.csrf_token,
            self.poll_cursor,
            self.notice_room,
        )

    async def insert(self) -> None:
        q = (
            'INSERT INTO "user" (mxid, twid, auth_token, csrf_token, poll_cursor, notice_room) '
            "VALUES ($1, $2, $3, $4, $5, $6)"
        )
        await self.db.execute(q, *self._values)

    async def update(self) -> None:
        q = (
            'UPDATE "user" SET twid=$2, auth_token=$3, csrf_token=$4,'
            "                  poll_cursor=$5, notice_room=$6 "
            "WHERE mxid=$1"
        )
        await self.db.execute(q, *self._values)

    @classmethod
    async def get_by_mxid(cls, mxid: UserID) -> User | None:
        q = (
            "SELECT mxid, twid, auth_token, csrf_token, poll_cursor, notice_room "
            'FROM "user" WHERE mxid=$1'
        )
        row = await cls.db.fetchrow(q, mxid)
        if not row:
            return None
        return cls(**row)

    @classmethod
    async def get_by_twid(cls, twid: int) -> User | None:
        q = (
            "SELECT mxid, twid, auth_token, csrf_token, poll_cursor, notice_room "
            'FROM "user" WHERE twid=$1'
        )
        row = await cls.db.fetchrow(q, twid)
        if not row:
            return None
        return cls(**row)

    @classmethod
    async def all_logged_in(cls) -> list[User]:
        q = (
            "SELECT mxid, twid, auth_token, csrf_token, poll_cursor, notice_room "
            'FROM "user" WHERE twid IS NOT NULL AND auth_token IS NOT NULL'
        )
        rows = await cls.db.fetch(q)
        return [cls(**row) for row in rows]



# mautrix-twitter - A Matrix-Twitter DM puppeting bridge
# Copyright (C) 2022 Tulir Asokan
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
from __future__ import annotations

from typing import TYPE_CHECKING

from mautrix.bridge import BaseMatrixHandler
from mautrix.types import (
    Event,
    EventID,
    EventType,
    PresenceEvent,
    ReactionEvent,
    ReactionEventContent,
    ReceiptEvent,
    RedactionEvent,
    RelationType,
    RoomID,
    SingleReceiptEventContent,
    TypingEvent,
    UserID,
)

from . import portal as po, user as u
from .db import Message as DBMessage, Reaction as DBReaction

if TYPE_CHECKING:
    from .__main__ import TwitterBridge


class MatrixHandler(BaseMatrixHandler):
    def __init__(self, bridge: "TwitterBridge") -> None:
        prefix, suffix = bridge.config["bridge.username_template"].format(userid=":").split(":")
        homeserver = bridge.config["homeserver.domain"]
        self.user_id_prefix = f"@{prefix}"
        self.user_id_suffix = f"{suffix}:{homeserver}"

        super().__init__(bridge=bridge)

    async def send_welcome_message(self, room_id: RoomID, inviter: u.User) -> None:
        await super().send_welcome_message(room_id, inviter)
        if not inviter.notice_room:
            inviter.notice_room = room_id
            await inviter.update()
            await self.az.intent.send_notice(
                room_id, "This room has been marked as your Twitter DM bridge notice room."
            )

    async def handle_leave(self, room_id: RoomID, user_id: UserID, event_id: EventID) -> None:
        portal = await po.Portal.get_by_mxid(room_id)
        if not portal:
            return

        user = await u.User.get_by_mxid(user_id, create=False)
        if not user:
            return

        await portal.handle_matrix_leave(user)

    @staticmethod
    async def handle_redaction(
        room_id: RoomID, user_id: UserID, event_id: EventID, redaction_event_id: EventID
    ) -> None:
        user = await u.User.get_by_mxid(user_id)
        if not user:
            return

        portal = await po.Portal.get_by_mxid(room_id)
        if not portal:
            return

        await portal.handle_matrix_redaction(user, event_id, redaction_event_id)

    @classmethod
    async def handle_reaction(
        cls,
        room_id: RoomID,
        user_id: UserID,
        event_id: EventID,
        content: ReactionEventContent,
    ) -> None:
        if content.relates_to.rel_type != RelationType.ANNOTATION:
            cls.log.debug(
                f"Ignoring m.reaction event in {room_id} from {user_id} with unexpected "
                f"relation type {content.relates_to.rel_type}"
            )
            return
        user = await u.User.get_by_mxid(user_id)
        if not user:
            return

        portal = await po.Portal.get_by_mxid(room_id)
        if not portal:
            return

        await portal.handle_matrix_reaction(
            user, event_id, content.relates_to.event_id, content.relates_to.key
        )

    async def handle_read_receipt(
        self, user: u.User, portal: po.Portal, event_id: EventID, data: SingleReceiptEventContent
    ) -> None:
        message = await DBMessage.get_by_mxid(event_id, portal.mxid)
        read_twid = 0
        if message:
            read_twid = message.twid
        else:
            # Can't find this message, perhaps it's a reaction?
            # Just grab the last thing we can find
            reaction = await DBReaction.get_last(portal.mxid)
            message = await DBMessage.get_last(portal.mxid)
            if message:
                read_twid = message.twid
            if reaction:
                read_twid = max(read_twid, reaction.tw_reaction_id)
        if read_twid == 0:
            user.log.debug(f"Unable to find message mxid:{event_id}, not marking as read")
            return
        user.log.debug(f"Marking messages in {portal.twid} read up to {read_twid}")
        await user.client.conversation(portal.twid).mark_read(read_twid)

    @staticmethod
    async def handle_typing(room_id: RoomID, typing: list[UserID]) -> None:
        # TODO implement
        pass

    async def handle_event(self, evt: Event) -> None:
        if evt.type == EventType.ROOM_REDACTION:
            evt: RedactionEvent
            await self.handle_redaction(evt.room_id, evt.sender, evt.redacts, evt.event_id)
        elif evt.type == EventType.REACTION:
            evt: ReactionEvent
            await self.handle_reaction(evt.room_id, evt.sender, evt.event_id, evt.content)

    async def handle_ephemeral_event(
        self, evt: ReceiptEvent | PresenceEvent | TypingEvent
    ) -> None:
        if evt.type == EventType.TYPING:
            await self.handle_typing(evt.room_id, evt.content.user_ids)
        else:
            await super().handle_ephemeral_event(evt)



# mautrix-twitter - A Matrix-Twitter DM puppeting bridge
# Copyright (C) 2022 Tulir Asokan
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
from __future__ import annotations

from typing import Awaitable
import json
import logging

from aiohttp import web

from mautrix.types import UserID
from mautrix.util.logging import TraceLogger
from mautwitdm.errors import TwitterError

from .. import user as u


class ProvisioningAPI:
    log: TraceLogger = logging.getLogger("mau.web.provisioning")
    app: web.Application

    def __init__(self, shared_secret: str) -> None:
        self.app = web.Application()
        self.shared_secret = shared_secret
        self.app.router.add_get("/api/whoami", self.status)
        self.app.router.add_options("/api/login", self.login_options)
        self.app.router.add_post("/api/login", self.login)
        self.app.router.add_post("/api/logout", self.logout)

    @property
    def _acao_headers(self) -> dict[str, str]:
        return {
            "Access-Control-Allow-Origin": "*",
            "Access-Control-Allow-Headers": "Authorization, Content-Type",
            "Access-Control-Allow-Methods": "POST, OPTIONS",
        }

    @property
    def _headers(self) -> dict[str, str]:
        return {
            **self._acao_headers,
            "Content-Type": "application/json",
        }

    async def login_options(self, _: web.Request) -> web.Response:
        return web.Response(status=200, headers=self._headers)

    def check_token(self, request: web.Request) -> Awaitable[u.User]:
        try:
            token = request.headers["Authorization"]
            token = token[len("Bearer ") :]
        except KeyError:
            raise web.HTTPBadRequest(
                body='{"error": "Missing Authorization header"}', headers=self._headers
            )
        except IndexError:
            raise web.HTTPBadRequest(
                body='{"error": "Malformed Authorization header"}',
                headers=self._headers,
            )
        if token != self.shared_secret:
            raise web.HTTPForbidden(body='{"error": "Invalid token"}', headers=self._headers)
        try:
            user_id = request.query["user_id"]
        except KeyError:
            raise web.HTTPBadRequest(
                body='{"error": "Missing user_id query param"}', headers=self._headers
            )

        return u.User.get_by_mxid(UserID(user_id))

    async def status(self, request: web.Request) -> web.Response:
        user = await self.check_token(request)
        data = {
            "permissions": user.permission_level,
            "mxid": user.mxid,
            "twitter": None,
        }
        if await user.is_logged_in():
            data["twitter"] = (await user.get_info()).serialize()
        return web.json_response(data, headers=self._acao_headers)

    async def login(self, request: web.Request) -> web.Response:
        user = await self.check_token(request)

        try:
            data = await request.json()
        except json.JSONDecodeError:
            raise web.HTTPBadRequest(body='{"error": "Malformed JSON"}', headers=self._headers)

        try:
            auth_token = data["auth_token"]
            csrf_token = data["csrf_token"]
        except KeyError:
            raise web.HTTPBadRequest(body='{"error": "Missing keys"}', headers=self._headers)

        try:
            await user.locked_connect(auth_token=auth_token, csrf_token=csrf_token)
        except TwitterError as e:
            self.log.debug("Failed to log in with TwitterError", exc_info=True)
            raise web.HTTPUnauthorized(
                body=json.dumps({"error": e.message}), headers=self._headers
            )
        except Exception:
            self.log.debug("Failed to log in with generic Exception", exc_info=True)
            raise web.HTTPUnauthorized(
                body='{"error": "Twitter authorization failed"}', headers=self._headers
            )
        return web.Response(body="{}", status=200, headers=self._headers)

    async def logout(self, request: web.Request) -> web.Response:
        user = await self.check_token(request)
        await user.logout()
        return web.json_response({}, headers=self._acao_headers)



# mautrix-twitter - A Matrix-Twitter DM puppeting bridge
# Copyright (C) 2022 Tulir Asokan
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
from __future__ import annotations

from typing import TYPE_CHECKING, ClassVar

from attr import dataclass
import asyncpg

from mautrix.types import RoomID
from mautrix.util.async_db import Database
from mautwitdm.types import ConversationType

fake_db = Database.create("") if TYPE_CHECKING else None


@dataclass
class Portal:
    db: ClassVar[Database] = fake_db

    twid: str
    receiver: int
    conv_type: ConversationType
    other_user: int | None
    mxid: RoomID | None
    name: str | None
    encrypted: bool
    next_batch_id: str | None

    @property
    def _values(self):
        return (
            self.twid,
            self.receiver,
            self.conv_type.value,
            self.other_user,
            self.mxid,
            self.name,
            self.encrypted,
            self.next_batch_id,
        )

    async def insert(self) -> None:
        q = (
            "INSERT INTO portal (twid, receiver, conv_type, other_user, mxid, name, encrypted, next_batch_id) "
            "VALUES ($1, $2, $3, $4, $5, $6, $7, $8)"
        )
        await self.db.execute(q, *self._values)

    async def update(self) -> None:
        q = (
            "UPDATE portal SET conv_type=$3, other_user=$4, mxid=$5, name=$6, encrypted=$7, next_batch_id=$8 "
            "WHERE twid=$1 AND receiver=$2"
        )
        await self.db.execute(q, *self._values)

    @classmethod
    def _from_row(cls, row: asyncpg.Record) -> "Portal":
        data = {**row}
        return cls(conv_type=ConversationType(data.pop("conv_type")), **data)

    @classmethod
    async def get_by_mxid(cls, mxid: RoomID) -> Portal | None:
        q = (
            "SELECT twid, receiver, conv_type, other_user, mxid, name, encrypted, next_batch_id "
            "FROM portal WHERE mxid=$1"
        )
        row = await cls.db.fetchrow(q, mxid)
        if not row:
            return None
        return cls._from_row(row)

    @classmethod
    async def get_by_twid(cls, twid: str, receiver: int = 0) -> Portal | None:
        q = (
            "SELECT twid, receiver, conv_type, other_user, mxid, name, encrypted, next_batch_id "
            "FROM portal WHERE twid=$1 AND receiver=$2"
        )
        row = await cls.db.fetchrow(q, twid, receiver)
        if not row:
            return None
        return cls._from_row(row)

    @classmethod
    async def find_private_chats_of(cls, receiver: int) -> list[Portal]:
        q = (
            "SELECT twid, receiver, conv_type, other_user, mxid, name, encrypted, next_batch_id FROM portal "
            "WHERE receiver=$1 AND conv_type='ONE_TO_ONE'"
        )
        rows = await cls.db.fetch(q, receiver)
        return [cls._from_row(row) for row in rows]

    @classmethod
    async def find_private_chats_with(cls, other_user: int) -> list[Portal]:
        q = (
            "SELECT twid, receiver, conv_type, other_user, mxid, name, encrypted, next_batch_id FROM portal "
            "WHERE other_user=$1 AND conv_type='ONE_TO_ONE'"
        )
        rows = await cls.db.fetch(q, other_user)
        return [cls._from_row(row) for row in rows]

    @classmethod
    async def all_with_room(cls) -> list[Portal]:
        q = (
            "SELECT twid, receiver, conv_type, other_user, mxid, name, encrypted, next_batch_id FROM portal "
            "WHERE mxid IS NOT NULL"
        )
        rows = await cls.db.fetch(q)
        return [cls._from_row(row) for row in rows]



# mautrix-twitter - A Matrix-Twitter DM puppeting bridge
# Copyright (C) 2022 Tulir Asokan
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
from __future__ import annotations

from typing import TYPE_CHECKING, Any, AsyncGenerator, Awaitable, Literal, NamedTuple, Tuple, cast
from collections import deque
from datetime import datetime, timedelta
import asyncio
import base64
import hashlib
import time

from yarl import URL
import magic

from mautrix.appservice import DOUBLE_PUPPET_SOURCE_KEY, AppService, IntentAPI
from mautrix.bridge import BasePortal, NotificationDisabler, async_getter_lock
from mautrix.errors import MatrixError, MForbidden
from mautrix.types import (
    AudioInfo,
    BatchSendEvent,
    BatchSendStateEvent,
    BeeperMessageStatusEventContent,
    ContentURI,
    EncryptedFile,
    EventID,
    EventType,
    ImageInfo,
    MediaMessageEventContent,
    Membership,
    MemberStateEventContent,
    MessageEventContent,
    MessageStatus,
    MessageStatusReason,
    MessageType,
    ReactionEventContent,
    RelatesTo,
    RelationType,
    RoomID,
    TextMessageEventContent,
    ThumbnailInfo,
    VideoInfo,
)
from mautrix.util import background_task, ffmpeg, variation_selector
from mautrix.util.message_send_checkpoint import MessageSendCheckpointStatus
from mautrix.util.simple_lock import SimpleLock
from mautwitdm.types import (
    Conversation,
    ConversationType,
    MessageData,
    MessageEntry,
    Participant,
    ReactionCreateEntry,
    VideoVariant,
)

from . import backfill as b, matrix as m, puppet as p, user as u
from .config import Config
from .db import (
    BackfillStatus as DBBackfillStatus,
    Message as DBMessage,
    Portal as DBPortal,
    Reaction as DBReaction,
)
from .formatter import twitter_to_matrix

if TYPE_CHECKING:
    from .__main__ import TwitterBridge

try:
    from mautrix.crypto.attachments import decrypt_attachment, encrypt_attachment
except ImportError:
    encrypt_attachment = decrypt_attachment = None

StateBridge = EventType.find("m.bridge", EventType.Class.STATE)
StateHalfShotBridge = EventType.find("uk.half-shot.bridge", EventType.Class.STATE)
StateMarker = EventType.find("org.matrix.msc2716.marker", EventType.Class.STATE)


class UnsupportedAttachmentError(NotImplementedError):
    pass


class ReuploadedMediaInfo(NamedTuple):
    mxc: ContentURI | None
    decryption_info: EncryptedFile | None
    mime_type: str
    file_name: str
    size: int


class Portal(DBPortal, BasePortal):
    by_mxid: dict[RoomID, Portal] = {}
    by_twid: dict[tuple[str, int], Portal] = {}
    config: Config
    matrix: "m.MatrixHandler"
    az: AppService
    private_chat_portal_meta: Literal["default", "always", "never"]

    _main_intent: IntentAPI | None
    _create_room_lock: asyncio.Lock
    backfill_lock: SimpleLock
    _msgid_dedup: deque[int]
    _reqid_dedup: set[str]
    _reaction_dedup: deque[tuple[int, int, str]]

    _main_intent: IntentAPI
    _last_participant_update: set[int]
    _reaction_lock: asyncio.Lock

    def __init__(
        self,
        twid: str,
        receiver: int,
        conv_type: ConversationType,
        other_user: int | None = None,
        mxid: RoomID | None = None,
        name: str | None = None,
        encrypted: bool = False,
        next_batch_id: str | None = None,
    ) -> None:
        super().__init__(
            twid,
            receiver,
            conv_type,
            other_user,
            mxid,
            name,
            encrypted,
            next_batch_id,
        )
        self._create_room_lock = asyncio.Lock()
        self.log = self.log.getChild(twid)
        self._msgid_dedup = deque(maxlen=100)
        self._reaction_dedup = deque(maxlen=100)
        self._reqid_dedup = set()
        self._last_participant_update = set()

        self.backfill_lock = SimpleLock(
            "Waiting for backfilling to finish before handling %s", log=self.log
        )
        self._main_intent = None
        self._reaction_lock = asyncio.Lock()

    @property
    def is_direct(self) -> bool:
        return self.conv_type == ConversationType.ONE_TO_ONE

    @property
    def set_dm_room_metadata(self) -> bool:
        return (
            not self.is_direct
            or self.private_chat_portal_meta == "always"
            or (self.encrypted and self.private_chat_portal_meta != "never")
        )

    @property
    def main_intent(self) -> IntentAPI:
        if not self._main_intent:
            raise ValueError("Portal must be postinit()ed before main_intent can be used")
        return self._main_intent

    @classmethod
    def init_cls(cls, bridge: "TwitterBridge") -> None:
        cls.config = bridge.config
        cls.matrix = bridge.matrix
        cls.az = bridge.az
        cls.loop = bridge.loop
        BasePortal.bridge = bridge
        cls.private_chat_portal_meta = cls.config["bridge.private_chat_portal_meta"]
        NotificationDisabler.puppet_cls = p.Puppet
        NotificationDisabler.config_enabled = cls.config["bridge.backfill.disable_notifications"]

    # region Misc

    async def _send_delivery_receipt(self, event_id: EventID) -> None:
        if event_id and self.config["bridge.delivery_receipts"]:
            try:
                await self.az.intent.mark_read(self.mxid, event_id)
            except Exception:
                self.log.exception("Failed to send delivery receipt for %s", event_id)

    async def _upsert_reaction(
        self,
        existing: DBReaction | None,
        intent: IntentAPI,
        mxid: EventID,
        message: DBMessage,
        sender: u.User | p.Puppet,
        reaction: str,
        reaction_id: int | None = None,
    ) -> None:
        if existing:
            self.log.debug(
                f"_upsert_reaction redacting {existing.mxid} and inserting {mxid}"
                f" (message: {message.mxid})"
            )
            await intent.redact(existing.mx_room, existing.mxid)
            await existing.edit(
                reaction=reaction, mxid=mxid, mx_room=message.mx_room, tw_reaction_id=reaction_id
            )
        else:
            self.log.debug(f"_upsert_reaction inserting {mxid} (message: {message.mxid})")
            await DBReaction(
                mxid=mxid,
                mx_room=message.mx_room,
                tw_msgid=message.twid,
                tw_receiver=self.receiver,
                tw_sender=sender.twid,
                reaction=reaction,
                tw_reaction_id=reaction_id,
            ).insert()

    # endregion
    # region Matrix event handling

    async def _send_bridge_success(
        self,
        sender: u.User,
        event_id: EventID,
        event_type: EventType,
        msgtype: MessageType | None = None,
    ) -> None:
        sender.send_remote_checkpoint(
            status=MessageSendCheckpointStatus.SUCCESS,
            event_id=event_id,
            room_id=self.mxid,
            event_type=event_type,
            message_type=msgtype,
        )
        background_task.create(self._send_message_status(event_id, err=None))
        await self._send_delivery_receipt(event_id)

    async def _send_bridge_error(
        self,
        sender: u.User,
        err: Exception,
        event_id: EventID,
        event_type: EventType,
        message_type: MessageType | None = None,
    ) -> None:
        sender.send_remote_checkpoint(
            self._status_from_exception(err),
            event_id,
            self.mxid,
            event_type,
            message_type=message_type,
            error=err,
        )

        send_notice = not isinstance(err, NotImplementedError)
        if self.config["bridge.delivery_error_reports"] and send_notice:
            event_type_str = {
                EventType.REACTION: "reaction",
                EventType.ROOM_REDACTION: "redaction",
            }.get(event_type, "message")
            await self._send_message(
                self.main_intent,
                TextMessageEventContent(
                    msgtype=MessageType.NOTICE,
                    body=f"\u26a0 Your {event_type_str} may not have been bridged: {str(err)}",
                ),
            )
        background_task.create(self._send_message_status(event_id, err))

    async def _send_message_status(self, event_id: EventID, err: Exception | None) -> None:
        if not self.config["bridge.message_status_events"]:
            return
        intent = self.az.intent if self.encrypted else self.main_intent
        status = BeeperMessageStatusEventContent(
            network=self.bridge_info_state_key,
            relates_to=RelatesTo(
                rel_type=RelationType.REFERENCE,
                event_id=event_id,
            ),
        )
        if err:
            status.status = MessageStatus.RETRIABLE
            status.reason = MessageStatusReason.GENERIC_ERROR
            status.error = str(err)
            if isinstance(err, NotImplementedError):
                if isinstance(err, UnsupportedAttachmentError):
                    status.message = str(err)
                status.status = MessageStatus.FAIL
                status.reason = MessageStatusReason.UNSUPPORTED
        else:
            status.status = MessageStatus.SUCCESS

        await intent.send_message_event(
            room_id=self.mxid,
            event_type=EventType.BEEPER_MESSAGE_STATUS,
            content=status,
        )

    @staticmethod
    def _status_from_exception(e: Exception) -> MessageSendCheckpointStatus:
        if isinstance(e, NotImplementedError):
            return MessageSendCheckpointStatus.UNSUPPORTED
        return MessageSendCheckpointStatus.PERM_FAILURE

    async def handle_matrix_message(
        self, sender: u.User, message: MessageEventContent, event_id: EventID
    ) -> None:
        try:
            await self._handle_matrix_message(sender, message, event_id)
        except Exception as e:
            self.log.exception(f"Error handling Matrix event {event_id}")
            await self._send_bridge_error(
                sender, e, event_id, EventType.ROOM_MESSAGE, message.msgtype
            )
        else:
            await self._send_bridge_success(
                sender, event_id, EventType.ROOM_MESSAGE, message.msgtype
            )

    async def _handle_matrix_message(
        self, sender: u.User, message: MessageEventContent, event_id: EventID
    ) -> None:
        if not sender.client:
            raise NotImplementedError("user is not connected")
        request_id = str(sender.client.new_request_id())
        self._reqid_dedup.add(request_id)
        media_id = None
        if message.msgtype == MessageType.TEXT:
            text = message.body
        elif message.msgtype == MessageType.EMOTE:
            text = f"/me {message.body}"
        elif message.msgtype.is_media:
            if message.msgtype != MessageType.IMAGE and message.msgtype != MessageType.VIDEO:
                raise UnsupportedAttachmentError(
                    "Non-image/video files are not supported by Twitter"
                )
            if message.file and decrypt_attachment:
                data = await self.main_intent.download_media(message.file.url)
                data = decrypt_attachment(
                    data,
                    message.file.key.key,
                    message.file.hashes.get("sha256"),
                    message.file.iv,
                )
            else:
                data = await self.main_intent.download_media(message.url)
            mime_type = message.info.mimetype or magic.from_buffer(data, mime=True)
            upload_resp = await sender.client.upload(data, mime_type=mime_type)
            media_id = upload_resp.media_id
            filename = message.get("filename", None)
            if filename and filename != message.body:
                text = message.body
            else:
                text = ""
        else:
            raise NotImplementedError(f"unsupported msgtype '{message.msgtype.value}'")
        reply_to = None
        if message.get_reply_to():
            reply_to_msg = await DBMessage.get_by_mxid(message.get_reply_to(), self.mxid)
            if reply_to_msg:
                reply_to = reply_to_msg.twid
        resp = await sender.client.conversation(self.twid).send(
            text, media_id=media_id, request_id=request_id, reply_to_id=reply_to
        )
        resp_msg_id = int(resp.entries[0].message.id)
        self._msgid_dedup.appendleft(resp_msg_id)
        msg = DBMessage(mxid=event_id, mx_room=self.mxid, twid=resp_msg_id, receiver=self.receiver)
        await msg.insert()
        self._reqid_dedup.remove(request_id)
        self.log.debug(f"Handled Matrix message {event_id} -> {resp_msg_id}")

    async def handle_matrix_reaction(
        self, sender: u.User, event_id: EventID, reacting_to: EventID, reaction: str
    ) -> None:
        try:
            await self._handle_matrix_reaction(sender, event_id, reacting_to, reaction)
        except Exception as e:
            self.log.exception(f"Failed to react to {event_id}")
            await self._send_bridge_error(sender, e, event_id, EventType.REACTION)
        else:
            await self._send_bridge_success(sender, event_id, EventType.REACTION)

    async def _handle_matrix_reaction(
        self, sender: u.User, event_id: EventID, reacting_to: EventID, reaction: str
    ) -> None:
        reaction = variation_selector.remove(reaction)
        message = await DBMessage.get_by_mxid(reacting_to, self.mxid)
        if not message:
            raise NotImplementedError(f"Unknown reaction target event")

        async with self._reaction_lock:
            dedup_id = (message.twid, sender.twid, reaction)
            self._reaction_dedup.appendleft(dedup_id)

            existing = await DBReaction.get_by_message_twid(
                message.twid, message.receiver, sender.twid
            )
            if existing and existing.reaction == reaction:
                return

            await sender.client.conversation(self.twid).react(message.twid, reaction)
            await self._upsert_reaction(
                existing, self.main_intent, event_id, message, sender, reaction
            )
            self.log.debug(f"{sender.mxid} reacted to {message.twid} with {reaction}")

    async def handle_matrix_redaction(
        self, sender: u.User, event_id: EventID, redaction_event_id: EventID
    ) -> None:
        try:
            await self._handle_matrix_redaction(sender, event_id)
        except Exception as e:
            self.log.exception(f"Error handling redaction of {event_id}")
            await self._send_bridge_error(sender, e, redaction_event_id, EventType.ROOM_REDACTION)
        else:
            await self._send_bridge_success(sender, redaction_event_id, EventType.ROOM_REDACTION)

    async def _handle_matrix_redaction(self, sender: u.User, event_id: EventID) -> None:
        async with self._reaction_lock:
            reaction = await DBReaction.get_by_mxid(event_id, self.mxid)
            if reaction:
                await reaction.delete()
                await sender.client.conversation(self.twid).delete_reaction(
                    reaction.tw_msgid, reaction.reaction
                )
                self.log.trace(f"Removed {reaction} after Matrix redaction")
                return

        raise NotImplementedError("Message redactions are not supported")

    async def handle_matrix_leave(self, user: u.User) -> None:
        if self.is_direct:
            self.log.info(f"{user.mxid} left private chat portal with {self.twid}")
            if user.twid == self.receiver:
                self.log.info(
                    f"{user.mxid} was the recipient of this portal. Cleaning up and deleting..."
                )
                await self.cleanup_and_delete()
        else:
            self.log.debug(f"{user.mxid} left portal to {self.twid}")
            # TODO cleanup if empty

    # endregion
    # region Twitter event handling

    async def handle_twitter_message(
        self, source: u.User, sender: p.Puppet, message: MessageData, request_id: str
    ) -> None:
        if await self._twitter_message_dedupe(request_id, int(message.id), message.sender_id):
            await self._handle_deduplicated_twitter_message(
                source, sender, message, int(message.id)
            )

    async def _twitter_message_dedupe(self, request_id: str, msg_id: int, sender_id: str) -> bool:
        if request_id in self._reqid_dedup:
            self.log.debug(
                f"Ignoring message {msg_id} by {sender_id}"
                " as it was sent by us (request_id in dedup queue)"
            )
            return False
        if msg_id in self._msgid_dedup:
            self.log.debug(
                f"Ignoring message {msg_id} by {sender_id}"
                " as it was already handled (message.id in dedup queue)"
            )
            return False
        self._msgid_dedup.appendleft(msg_id)

        if await DBMessage.get_by_twid(msg_id, self.receiver) is not None:
            self.log.debug(
                f"Ignoring message {msg_id} by {sender_id}"
                " as it was already handled (message.id found in database)"
            )
            return False

        return True

    def deterministic_event_id(self, msg_id: str, part: str) -> EventID:
        hash_content = f"{self.mxid}/twitter/{msg_id}/{part}"
        hashed = hashlib.sha256(hash_content.encode("utf-8")).digest()
        b64hash = base64.urlsafe_b64encode(hashed).decode("utf-8").rstrip("=")
        return EventID(f"${b64hash}:twitter.com")

    async def _convert_twitter_message(
        self, source: u.User, sender: p.Puppet, message: MessageData
    ) -> list[MessageEventContent]:
        converted = []
        intent = sender.intent_for(self)
        if message.reply_data:
            reply_to_msg = await DBMessage.get_by_twid(int(message.reply_data.id), self.receiver)
        else:
            reply_to_msg = None
        media_content = None
        if message.attachment and message.attachment.media:
            media_content = await self._handle_twitter_media(source, intent, message)
            if media_content:
                if reply_to_msg:
                    media_content.set_reply(reply_to_msg.mxid)
                converted.append(media_content)
        if message.text and not message.text.isspace():
            text_content = await twitter_to_matrix(message)
            text_content["com.beeper.linkpreviews"] = await self._twitter_preview_to_beeper(
                source, intent, message
            )
            if reply_to_msg:
                text_content.set_reply(reply_to_msg.mxid)
            if media_content and self.config["bridge.caption_in_message"]:
                media_content["filename"] = media_content.body
                media_content.body = text_content.body
                if text_content.formatted_body:
                    media_content["format"] = str(text_content.format)
                    media_content["formatted_body"] = text_content.formatted_body
            else:
                converted.append(text_content)
        return converted

    async def _handle_deduplicated_twitter_message(
        self, source: u.User, sender: p.Puppet, message: MessageData, msg_id: int
    ) -> None:
        intent = sender.intent_for(self)
        event_id = None
        converted = await self._convert_twitter_message(source, sender, message)
        for content in converted:
            event_id = await self._send_message(intent, content, timestamp=message.time)
        if event_id:
            msg = DBMessage(
                mxid=event_id,
                mx_room=self.mxid,
                twid=msg_id,
                receiver=self.receiver,
            )
            await msg.insert()
            await self._send_delivery_receipt(event_id)
            self.log.debug(f"Handled Twitter message {msg_id} -> {event_id}")

    @staticmethod
    def _is_better_mime(best: VideoVariant, current: VideoVariant) -> bool:
        order = ["video/mp4"]
        try:
            best_quality = order.index(best.content_type)
        except (IndexError, ValueError):
            best_quality = -1
        try:
            current_quality = order.index(current.content_type)
        except (IndexError, ValueError):
            current_quality = -1
        return current_quality > best_quality

    async def _twitter_preview_to_beeper(
        self, source: u.User, intent: IntentAPI, message: MessageData
    ) -> list[dict[str, Any]]:
        if not message.attachment or not message.attachment.url_preview:
            return []
        if message.attachment.card:
            preview = await self._twitter_card_to_beeper(source, intent, message)
        elif message.attachment.tweet:
            preview = await self._twitter_tweet_to_beeper(source, intent, message)
        else:
            return []
        return [{k: v for k, v in preview.items() if v is not None}]

    async def _twitter_card_to_beeper(
        self, source: u.User, intent: IntentAPI, message: MessageData
    ) -> dict[str, Any]:
        card = message.attachment.card
        card_url = card.string("card_url", "")
        if card_url.startswith("https://t.co") and message.entities:
            try:
                # Try to find the actual expanded URL from the message entities
                card_url = next(e for e in message.entities.urls if e.url == card_url).expanded_url
            except StopIteration:
                pass
        preview = {
            "matched_url": card_url,
            "og:title": card.string("title"),
            "og:description": card.string("description"),
        }
        image = (
            card.image("photo_image_full_size_original")
            or card.image("summary_photo_image_original")
            or card.image("thumbnail_image_original")
        )
        if image:
            preview.update(
                {
                    **await self._twitter_image_to_beeper(source, intent, image.url),
                    "og:image:width": image.width,
                    "og:image:height": image.height,
                }
            )
        return preview

    async def _twitter_tweet_to_beeper(
        self, source: u.User, intent: IntentAPI, message: MessageData
    ) -> dict[str, Any]:
        tweet = message.attachment.tweet.status
        preview = {
            "matched_url": message.attachment.tweet.expanded_url,
            "og:url": message.attachment.tweet.expanded_url,
            "og:title": f"{tweet.user.name} on Twitter",
            "og:description": tweet.full_text,
        }
        if tweet.extended_entities and tweet.extended_entities.media:
            media = tweet.extended_entities.media[0]
            if "medium" in media.sizes:
                size_name = "medium"
            else:
                size_name = next(media.sizes.keys().__iter__())
            url = str(URL(media.media_url_https).update_query({"name": size_name}))
            preview.update(
                {
                    **await self._twitter_image_to_beeper(source, intent, url),
                    "og:image:width": media.sizes[size_name].w,
                    "og:image:height": media.sizes[size_name].h,
                }
            )

        return preview

    async def _twitter_image_to_beeper(
        self, source: u.User, intent: IntentAPI, url: str
    ) -> dict[str, Any]:
        info = await self._reupload_twitter_media(source, url, intent)
        output = {
            "og:image:type": info.mime_type,
            "matrix:image:size": info.size,
        }
        if info.decryption_info:
            output["beeper:image:encryption"] = info.decryption_info.serialize()
        else:
            output["og:image"] = info.mxc
        return output

    async def _handle_twitter_media(
        self, source: u.User, intent: IntentAPI, message: MessageData
    ) -> MediaMessageEventContent | None:
        media = message.attachment.media
        reuploaded_info = await self._reupload_twitter_media(source, media.media_url_https, intent)
        thumbnail_info = None
        if media.video_info:
            thumbnail_info = reuploaded_info
            best_variant = None
            for variant in media.video_info.variants:
                if (
                    not best_variant
                    or (variant.bitrate or 0) > (best_variant.bitrate or 0)
                    or self._is_better_mime(best_variant, variant)
                ):
                    best_variant = variant
            reuploaded_info = await self._reupload_twitter_media(
                source, best_variant.url, intent, convert_to_audio=media.audio_only
            )
        content = MediaMessageEventContent(
            body=reuploaded_info.file_name,
            url=reuploaded_info.mxc,
            file=reuploaded_info.decryption_info,
            external_url=media.media_url_https,
        )
        if message.attachment.video and message.attachment.video.audio_only:
            content.msgtype = MessageType.AUDIO
            content.info = AudioInfo(
                mimetype=reuploaded_info.mime_type,
                size=reuploaded_info.size,
                duration=media.video_info.duration_millis or None,
            )
            content["org.matrix.msc1767.audio"] = (
                {
                    "duration": content.info.duration,
                }
                if content.info.duration
                else {}
            )
            content["org.matrix.msc3245.voice"] = {}
        elif message.attachment.video or (
            message.attachment.animated_gif and reuploaded_info.mime_type.startswith("video/")
        ):
            content.msgtype = MessageType.VIDEO
            content.info = VideoInfo(
                mimetype=reuploaded_info.mime_type,
                size=reuploaded_info.size,
                width=media.original_info.width,
                height=media.original_info.height,
                duration=media.video_info.duration_millis or None,
            )
            if message.attachment.animated_gif:
                content.info["fi.mau.loop"] = True
                content.info["fi.mau.autoplay"] = True
                content.info["fi.mau.hide_controls"] = True
                content.info["fi.mau.no_audio"] = True
        elif message.attachment.photo or message.attachment.animated_gif:
            content.msgtype = MessageType.IMAGE
            content.info = ImageInfo(
                mimetype=reuploaded_info.mime_type,
                size=reuploaded_info.size,
                width=media.original_info.width,
                height=media.original_info.height,
            )
        if thumbnail_info:
            content.info.thumbnail_url = thumbnail_info.mxc
            content.info.thumbnail_file = thumbnail_info.decryption_info
            content.info.thumbnail_info = ThumbnailInfo(
                mimetype=thumbnail_info.mime_type,
                size=thumbnail_info.size,
                width=media.original_info.width,
                height=media.original_info.height,
            )
        # Remove the attachment link from message.text
        start, end = media.indices
        message.text = message.text[:start] + message.text[end:]
        if message.entities and message.entities.urls:
            message.entities.urls = [u for u in message.entities.urls if u.url != media.url]
        return content

    async def _reupload_twitter_media(
        self, source: u.User, url: str, intent: IntentAPI, convert_to_audio: bool = False
    ) -> ReuploadedMediaInfo:
        file_name = URL(url).name
        data, mime_type = await source.client.download_media(url)

        if convert_to_audio and (mime_type.startswith("video/") or mime_type.startswith("audio/")):
            data = await ffmpeg.convert_bytes(
                data,
                ".ogg",
                output_args=("-c:a", "libopus"),
                input_mime=mime_type,
            )
            mime_type = "audio/ogg"
            file_name += ".ogg"

        upload_mime_type = mime_type
        upload_file_name = file_name
        decryption_info = None
        if self.encrypted and encrypt_attachment:
            data, decryption_info = encrypt_attachment(data)
            upload_mime_type = "application/octet-stream"
            upload_file_name = None

        mxc = await intent.upload_media(
            data,
            mime_type=upload_mime_type,
            filename=upload_file_name,
            async_upload=self.config["homeserver.async_media"],
        )

        if decryption_info:
            decryption_info.url = mxc
            mxc = None

        return ReuploadedMediaInfo(mxc, decryption_info, mime_type, file_name, len(data))

    async def handle_twitter_reaction_add(
        self,
        sender: p.Puppet,
        msg_id: int,
        emoji: str,
        time: datetime,
        reaction_id: int,
    ) -> None:
        async with self._reaction_lock:
            # TODO update the database with the reaction_id of outgoing reactions
            dedup_id = (msg_id, sender.twid, emoji)
            if dedup_id in self._reaction_dedup:
                self.log.debug(
                    f"Ignoring duplicate reaction from {sender.twid} to {msg_id} (dedup queue)"
                )
                return
            self._reaction_dedup.appendleft(dedup_id)

        existing = await DBReaction.get_by_message_twid(msg_id, self.receiver, sender.twid)
        if existing and existing.reaction == emoji:
            if not existing.tw_reaction_id:
                await existing.update_id(reaction_id)
            self.log.debug(
                f"Ignoring duplicate reaction from {sender.twid} to {msg_id} (database)"
            )
            return

        message = await DBMessage.get_by_twid(msg_id, self.receiver)
        if not message:
            self.log.debug(f"Ignoring reaction to unknown message {msg_id}")
            return

        intent = sender.intent_for(self)
        mxid = await intent.react(
            message.mx_room, message.mxid, variation_selector.add(emoji), timestamp=time
        )
        self.log.debug(f"{sender.twid} reacted to {msg_id}/{message.mxid} -> {mxid}")
        await self._upsert_reaction(existing, intent, mxid, message, sender, emoji, reaction_id)

    async def handle_twitter_reaction_remove(
        self, sender: p.Puppet, msg_id: int, emoji: str
    ) -> None:
        reaction = await DBReaction.get_by_message_twid(msg_id, self.receiver, sender.twid)
        if reaction and reaction.reaction == emoji:
            try:
                self._reaction_dedup.remove((msg_id, sender.twid, emoji))
            except ValueError:
                pass
            try:
                await sender.intent_for(self).redact(reaction.mx_room, reaction.mxid)
            except MForbidden:
                await self.main_intent.redact(reaction.mx_room, reaction.mxid)
            await reaction.delete()
            self.log.debug(f"Removed {reaction} after Twitter removal")

    async def handle_twitter_receipt(
        self, sender: p.Puppet, read_up_to: int, historical: bool = False
    ) -> None:
        message = await DBMessage.get_by_twid(read_up_to, self.receiver)
        if not message:
            message = await DBReaction.get_by_reaction_twid(read_up_to, self.receiver)
            if not message:
                self.log.debug(
                    f"Ignoring read receipt from {sender.twid} "
                    f"up to unknown message {read_up_to} ({historical=})"
                )
                return

        self.log.debug(
            f"{sender.twid} read messages up to {read_up_to} ({message.mxid}, {historical=})"
        )
        await sender.intent_for(self).mark_read(message.mx_room, message.mxid)

    # endregion
    # region Updating portal info

    async def update_info(self, conv: Conversation) -> None:
        if self.conv_type == ConversationType.ONE_TO_ONE:
            if not self.other_user:
                if len(conv.participants) == 1:
                    self.other_user = int(conv.participants[0].user_id)
                else:
                    participant = next(
                        pcp for pcp in conv.participants if int(pcp.user_id) != self.receiver
                    )
                    self.other_user = int(participant.user_id)
                await self.update()
            puppet = await p.Puppet.get_by_twid(self.other_user)
            if not self._main_intent:
                self._main_intent = puppet.default_mxid_intent
            changed = await self._update_name(puppet.name)
        else:
            changed = await self._update_name(conv.name)
        if changed:
            await self.update_bridge_info()
            await self.update()
        await self._update_participants(conv.participants)

    async def update_name(self, name: str) -> None:
        changed = await self._update_name(name)

        if changed:
            await self.update_bridge_info()
            await self.update()

    async def _update_name(self, name: str) -> bool:
        if self.name != name:
            self.name = name
            if self.mxid and self.set_dm_room_metadata:
                await self.main_intent.set_room_name(self.mxid, name)
            return True
        return False

    async def _update_participants(self, participants: list[Participant]) -> None:
        if not self.mxid:
            return

        # Store the current member list to prevent unnecessary updates
        current_members = {int(participant.user_id) for participant in participants}
        if current_members == self._last_participant_update:
            self.log.trace("Not updating participants: list matches cached list")
            return
        self._last_participant_update = current_members

        # Make sure puppets who should be here are here
        for participant in participants:
            twid = int(participant.user_id)
            puppet = await p.Puppet.get_by_twid(twid)
            await puppet.intent_for(self).ensure_joined(self.mxid, bot=self.main_intent)
            if participant.last_read_event_id:
                await self.handle_twitter_receipt(
                    puppet, int(participant.last_read_event_id), historical=True
                )

        # Kick puppets who shouldn't be here
        for user_id in await self.main_intent.get_room_members(self.mxid):
            twid = p.Puppet.get_id_from_mxid(user_id)
            if twid and twid not in current_members:
                await self.main_intent.kick_user(
                    self.mxid,
                    p.Puppet.get_mxid_from_id(twid),
                    reason="User had left this Twitter chat",
                )

    # endregion
    # region Backfilling

    async def backfill(self, source: u.User, is_initial: bool = False) -> int:
        limit = self.config["bridge.backfill.initial_limit"]
        if limit == 0:
            return 0
        elif limit < 0:
            limit = None
        with self.backfill_lock:
            return await self._backfill(source, limit, is_initial)

    async def _backfill(self, source: u.User, limit: int, is_initial: bool = False) -> int:
        if is_initial:
            self.log.debug("Backfilling initial batch through %s", source.mxid)
        elif not self.config["bridge.backfill.backwards"]:
            self.log.debug("Not backfilling history, disabled in config")
            return 0
        else:
            self.log.debug("Backfilling history through %s", source.mxid)

        first_message = await DBMessage.get_first(self.mxid)
        max_id = None
        if not is_initial:
            if first_message is not None:
                max_id = first_message.twid
            else:
                self.log.warn("Can't backfill without a first bridged message!")
                raise b.NoFirstMessageException()

        mark_read, entries = await self._fetch_backfill_entries(source, limit, max_id)
        if not entries:
            self.log.debug("Didn't get any entries from server")
            return 0

        self.log.debug("Got %d entries from server", len(entries))

        filled = 0
        if self.config["bridge.backfill.backwards"]:
            filled = await self._batch_handle_backfill(
                source, reversed(entries), is_initial, mark_read
            )
        else:
            backfill_leave = await self._invite_own_puppet_backfill(source)
            async with NotificationDisabler(self.mxid, source):
                for entry in reversed(entries):
                    await self._handle_backfill_entry(source, entry)
                    filled += 1
            for intent in backfill_leave:
                self.log.trace("Leaving room with %s post-backfill", intent.mxid)
                await intent.leave_room(self.mxid)
        self.log.info(
            "Backfilled %d messages (%d events) through %s", len(entries), filled, source.mxid
        )
        return filled

    async def _fetch_backfill_entries(
        self, source: u.User, limit: int, max_id: int | None = None
    ) -> tuple[bool, list[MessageEntry | ReactionCreateEntry]]:
        conv = source.client.conversation(self.twid)
        entries: list[MessageEntry | ReactionCreateEntry] = []
        message_count = 0
        self.log.debug("Fetching up to %d messages through %s", limit, source.twid)
        try:
            mark_read = False
            self.log.debug("Fetching with max_id %s", max_id)
            resp = await conv.fetch(max_id=max_id)
            resp_entries = (
                list(filter(lambda x: x is not None, resp.entries))
                if resp.entries is not None
                else None
            )
            try:
                if datetime.now() - timedelta(days=30) > datetime.fromtimestamp(
                    resp.conversations[self.twid].sort_timestamp
                ) or (
                    resp_entries is not None
                    and len(resp_entries) != 0
                    and int(resp.conversations[self.twid].last_read_event_id)
                    >= int(resp_entries[0].message.id)
                ):
                    mark_read = True
            except:
                mark_read = True

            while True:
                if resp_entries is None or len(resp_entries) == 0:
                    break
                for entry in resp_entries:
                    if entry and entry.message:
                        entries.append(entry.message)
                        message_count += 1
                        if entry.message.message_reactions:
                            entries += entry.message.message_reactions
                    if message_count >= limit:
                        break
                if message_count >= limit:
                    self.log.debug("Got more messages than limit")
                    break

                if resp.min_entry_id is None or resp.min_entry_id == max_id:
                    break
                max_id = resp.min_entry_id
                self.log.debug("Fetching more entries with max_id %s", max_id)
                resp = await conv.fetch(max_id=max_id)

            if len(entries) == 0:
                return None, None
            entries.sort(key=lambda ent: (ent.time, ent.id), reverse=True)
            self.log.debug("Finished fetching entries")
            return mark_read, entries
        except Exception:
            self.log.warning("Exception while fetching messages", exc_info=True)
            return None, None

    async def _invite_own_puppet_backfill(self, source: u.User) -> set[IntentAPI]:
        backfill_leave = set()
        # TODO we should probably only invite the puppet when needed
        if self.config["bridge.backfill.invite_own_puppet"]:
            self.log.debug("Adding %s's default puppet to room for backfilling", source.mxid)
            sender = await p.Puppet.get_by_twid(source.twid)
            await self.main_intent.invite_user(self.mxid, sender.default_mxid)
            await sender.default_mxid_intent.join_room_by_id(self.mxid)
            backfill_leave.add(sender.default_mxid_intent)
        return backfill_leave

    async def _handle_backfill_entry(
        self, source: u.User, entry: MessageEntry | ReactionCreateEntry
    ) -> None:
        sender = await p.Puppet.get_by_twid(int(entry.sender_id))
        if isinstance(entry, MessageEntry):
            await self.handle_twitter_message(source, sender, entry.message_data, entry.request_id)
        if isinstance(entry, ReactionCreateEntry):
            await self.handle_twitter_reaction_add(
                sender, int(entry.message_id), entry.reaction_emoji, entry.time, int(entry.id)
            )

    async def _batch_handle_backfill(
        self,
        source: u.User,
        entries: list[MessageEntry | ReactionCreateEntry],
        is_forward: bool,
        mark_read: bool,
    ) -> int:
        events = []
        twids = []
        users_in_batch = set()
        self.log.debug("Converting Twitter messages in batch")
        for i, entry in enumerate(entries):
            sender = await p.Puppet.get_by_twid(int(entry.sender_id))
            users_in_batch.add(sender.mxid)
            if isinstance(entry, MessageEntry):
                msg_id = int(entry.message_data.id)
                if await self._twitter_message_dedupe(
                    entry.request_id, msg_id, entry.message_data.sender_id
                ):
                    converted = await self._convert_twitter_message(
                        source, sender, entry.message_data
                    )
                    for i, content in enumerate(converted):
                        event_type = EventType.ROOM_MESSAGE
                        if self.encrypted and self.matrix.e2ee:
                            event_type, content = await self.matrix.e2ee.encrypt(
                                self.mxid, event_type, content
                            )
                        content[DOUBLE_PUPPET_SOURCE_KEY] = self.bridge.name
                        e = BatchSendEvent(
                            type=event_type,
                            content=content,
                            sender=sender.mxid,
                            timestamp=int(entry.message_data.time.timestamp() * 1000),
                        )
                        if self.bridge.homeserver_software.is_hungry:
                            e.event_id = self.deterministic_event_id(
                                entry.id, "main" if i + 1 == len(converted) else str(i)
                            )
                        events.append(e)
                        twids.append((msg_id, "message"))
            elif (
                isinstance(entry, ReactionCreateEntry)
                and self.bridge.homeserver_software.is_hungry
            ):
                content = ReactionEventContent(
                    relates_to=RelatesTo(
                        rel_type=RelationType.ANNOTATION,
                        event_id=self.deterministic_event_id(entry.message_id, "main"),
                        key=entry.reaction_emoji,
                    )
                )
                content[DOUBLE_PUPPET_SOURCE_KEY] = self.bridge.name
                e = BatchSendEvent(
                    type=EventType.REACTION,
                    content=content,
                    sender=sender.mxid,
                    event_id=self.deterministic_event_id(entry.id, f"reaction{i}"),
                    timestamp=int(entry.time.timestamp() * 1000),
                )
                events.append(e)
                twids.append(
                    (entry.id, "reaction", entry.message_id, entry.sender_id, entry.reaction_emoji)
                )
        if len(events) == 0:
            self.log.warn("No bridgeable messages in backfill batch")
            return 0

        intent = self.main_intent

        state_events_at_start = []
        if not self.bridge.homeserver_software.is_hungry:
            before_first_message_timestamp = events[0].timestamp - 1
            self.log.debug("Adding member state events to batch")
            for u in users_in_batch:
                puppet = await self.bridge.get_puppet(u)
                if puppet is None:
                    self.log.warn("No puppet found for user %s while backfilling!", u)
                    continue
                state_events_at_start.append(
                    BatchSendStateEvent(
                        type=EventType.ROOM_MEMBER,
                        content=MemberStateEventContent(
                            Membership.INVITE, avatar_url=puppet.photo_mxc, displayname=puppet.name
                        ),
                        sender=intent.mxid,
                        timestamp=before_first_message_timestamp,
                        state_key=u,
                    )
                )
                state_events_at_start.append(
                    BatchSendStateEvent(
                        type=EventType.ROOM_MEMBER,
                        content=MemberStateEventContent(
                            Membership.JOIN, avatar_url=puppet.photo_mxc, displayname=puppet.name
                        ),
                        sender=u,
                        timestamp=before_first_message_timestamp,
                        state_key=u,
                    )
                )

        first_event = None
        if not is_forward:
            first_event = await DBMessage.get_first(self.mxid)
            if first_event is None:
                raise b.NoFirstMessageException
            else:
                first_event = first_event.mxid
        self.log.debug("Sending batch send request")
        resp = await intent.batch_send(
            self.mxid,
            first_event,
            batch_id=None if is_forward else self.next_batch_id,
            events=events,
            state_events_at_start=state_events_at_start,
            beeper_new_messages=is_forward,
            beeper_mark_read_by=source.mxid if mark_read else None,
        )
        if resp.base_insertion_event_id is not None:
            self.log.debug("Sending msc2716 insertion marker event")
            await self.main_intent.send_state_event(
                self.mxid,
                StateMarker,
                {
                    "org.matrix.msc2716.marker.insertion": resp.base_insertion_event_id,
                    "com.beeper.timestamp": int(time.time() * 1000),
                },
                state_key=resp.base_insertion_event_id,
            )

        for i, event_id in enumerate(resp.event_ids):
            if twids[i][1] == "message":
                msg = DBMessage(event_id, self.mxid, twids[i][0], self.receiver)
                await msg.upsert()
            elif twids[i][1] == "reaction" and self.bridge.homeserver_software.is_hungry:
                reaction = DBReaction(
                    mxid=event_id,
                    mx_room=self.mxid,
                    tw_msgid=twids[i][2],
                    tw_receiver=self.receiver,
                    tw_sender=twids[i][3],
                    reaction=twids[i][4],
                    tw_reaction_id=twids[i][0],
                )
                await reaction.insert()
        self.next_batch_id = resp.next_batch_id
        await self.update()

        return len(events)

    # endregion
    # region Bridge info state event

    @property
    def bridge_info_state_key(self) -> str:
        return f"net.maunium.twitter://twitter/{self.twid}"

    @property
    def bridge_info(self) -> dict[str, Any]:
        return {
            "bridgebot": self.az.bot_mxid,
            "creator": self.main_intent.mxid,
            "protocol": {
                "id": "twitter",
                "displayname": "Twitter DM",
                "avatar_url": self.config["appservice.bot_avatar"],
            },
            "channel": {
                "id": self.twid,
                "displayname": self.name,
            },
        }

    async def update_bridge_info(self) -> None:
        if not self.mxid:
            self.log.debug("Not updating bridge info: no Matrix room created")
            return
        try:
            self.log.debug("Updating bridge info...")
            await self.main_intent.send_state_event(
                self.mxid, StateBridge, self.bridge_info, self.bridge_info_state_key
            )
            # TODO remove this once https://github.com/matrix-org/matrix-doc/pull/2346 is in spec
            await self.main_intent.send_state_event(
                self.mxid,
                StateHalfShotBridge,
                self.bridge_info,
                self.bridge_info_state_key,
            )
        except Exception:
            self.log.warning("Failed to update bridge info", exc_info=True)

    # endregion
    # region Creating Matrix rooms

    async def create_matrix_room(self, source: u.User, info: Conversation) -> RoomID | None:
        if self.mxid:
            try:
                await self._update_matrix_room(source, info)
            except Exception:
                self.log.exception("Failed to update portal")
            return self.mxid
        async with self._create_room_lock:
            return await self._create_matrix_room(source, info)

    def _get_invite_content(self, double_puppet: p.Puppet | None) -> dict[str, Any]:
        invite_content = {}
        if double_puppet:
            invite_content["fi.mau.will_auto_accept"] = True
        if self.is_direct:
            invite_content["is_direct"] = True
        return invite_content

    async def _add_user(self, user: u.User) -> None:
        puppet = await p.Puppet.get_by_custom_mxid(user.mxid)
        await self.main_intent.invite_user(
            self.mxid,
            user.mxid,
            check_cache=True,
            extra_content=self._get_invite_content(puppet),
        )
        if puppet:
            did_join = await puppet.intent.ensure_joined(self.mxid)
            if did_join and self.is_direct:
                await user.update_direct_chats({self.main_intent.mxid: [self.mxid]})

    async def _update_matrix_room(self, source: u.User, info: Conversation) -> None:
        await self._add_user(source)
        await self.update_info(info)

    async def _create_matrix_room(self, source: u.User, info: Conversation) -> RoomID | None:
        if self.mxid:
            await self._update_matrix_room(source, info)
            return self.mxid
        await self.update_info(info)
        self.log.debug("Creating Matrix room")
        initial_state = [
            {
                "type": str(StateBridge),
                "state_key": self.bridge_info_state_key,
                "content": self.bridge_info,
            },
            {
                # TODO remove this once https://github.com/matrix-org/matrix-doc/pull/2346 is in spec
                "type": str(StateHalfShotBridge),
                "state_key": self.bridge_info_state_key,
                "content": self.bridge_info,
            },
        ]
        invites = []
        if self.config["bridge.encryption.default"] and self.matrix.e2ee:
            self.encrypted = True
            initial_state.append(
                {
                    "type": "m.room.encryption",
                    "content": self.get_encryption_state_event_json(),
                }
            )
            if self.is_direct:
                invites.append(self.az.bot_mxid)

        # We lock backfill lock here so any messages that come between the room being created
        # and the initial backfill finishing wouldn't be bridged before the backfill messages.
        with self.backfill_lock:
            creation_content = {}
            if not self.config["bridge.federate_rooms"]:
                creation_content["m.federate"] = False
            self.mxid = await self.main_intent.create_room(
                name=self.name if self.set_dm_room_metadata else None,
                is_direct=self.is_direct,
                initial_state=initial_state,
                invitees=invites,
                creation_content=creation_content,
            )
            if not self.mxid:
                raise Exception("Failed to create room: no mxid returned")

            if self.encrypted and self.matrix.e2ee and self.is_direct:
                try:
                    await self.az.intent.ensure_joined(self.mxid)
                except Exception:
                    self.log.warning(f"Failed to add bridge bot to new private chat {self.mxid}")

            await self.update()
            self.log.debug(f"Matrix room created: {self.mxid}")
            self.by_mxid[self.mxid] = self
            await self._add_user(source)
            await self._update_participants(info.participants)

            puppet = await p.Puppet.get_by_custom_mxid(source.mxid)
            if puppet:
                try:
                    await puppet.intent.join_room_by_id(self.mxid)
                    if self.is_direct:
                        await source.update_direct_chats({self.main_intent.mxid: [self.mxid]})
                    if self.config["bridge.low_quality_tag"]:
                        await source.tag_room(
                            puppet,
                            self,
                            self.config["bridge.low_quality_tag"],
                            info.low_quality == True,
                        )
                    if self.config["bridge.low_quality_mute"]:
                        await source.set_muted(puppet, self, info.low_quality == True)
                except MatrixError:
                    self.log.debug(
                        "Failed to join custom puppet into newly created portal", exc_info=True
                    )

            if not info.trusted:
                msg = "This is a message request. Replying here will accept the request."
                if info.low_quality:
                    msg += ' Note: Twitter has marked this as a "low quality" message.'
                await self.main_intent.send_notice(self.mxid, msg)

            try:
                await self._enqueue_backfills(source)
            except Exception:
                self.log.exception("Failed to backfill new portal")

            # Update participants again after backfill to sync read receipts
            self._last_participant_update = set()
            await self._update_participants(info.participants)

        return self.mxid

    async def _enqueue_backfills(self, source: u.User) -> None:
        state = DBBackfillStatus(self.twid, self.receiver, source.twid, False, 0, 0)
        await state.insert()
        await b.BackfillStatus.recheck()

    # endregion
    # region Database getters

    async def postinit(self) -> None:
        self.by_twid[(self.twid, self.receiver)] = self
        if self.mxid:
            self.by_mxid[self.mxid] = self
        if self.other_user and self.is_direct:
            self._main_intent = (await p.Puppet.get_by_twid(self.other_user)).default_mxid_intent
        elif not self.is_direct:
            self._main_intent = self.az.intent

    async def delete(self) -> None:
        await DBMessage.delete_all(self.mxid)
        self.by_mxid.pop(self.mxid, None)
        self.mxid = None
        self.encrypted = False
        await self.update()

    async def save(self) -> None:
        await self.update()

    @classmethod
    def all_with_room(cls) -> AsyncGenerator[Portal, None]:
        return cls._db_to_portals(super().all_with_room())

    @classmethod
    def find_private_chats_with(cls, other_user: int) -> AsyncGenerator[Portal, None]:
        return cls._db_to_portals(super().find_private_chats_with(other_user))

    @classmethod
    async def _db_to_portals(cls, query: Awaitable[list[Portal]]) -> AsyncGenerator[Portal, None]:
        portals = await query
        for index, portal in enumerate(portals):
            try:
                yield cls.by_twid[(portal.twid, portal.receiver)]
            except KeyError:
                await portal.postinit()
                yield portal

    @classmethod
    @async_getter_lock
    async def get_by_mxid(cls, mxid: RoomID) -> Portal | None:
        try:
            return cls.by_mxid[mxid]
        except KeyError:
            pass

        portal = cast(cls, await super().get_by_mxid(mxid))
        if portal is not None:
            await portal.postinit()
            return portal

        return None

    @classmethod
    @async_getter_lock
    async def get_by_twid(
        cls,
        twid: str,
        *,
        receiver: int = 0,
        conv_type: ConversationType | None = None,
    ) -> Portal | None:
        if conv_type == ConversationType.GROUP_DM and receiver != 0:
            receiver = 0
        try:
            return cls.by_twid[(twid, receiver)]
        except KeyError:
            pass

        portal = cast(cls, await super().get_by_twid(twid, receiver))
        if portal is not None:
            await portal.postinit()
            return portal

        if conv_type is not None:
            portal = cls(twid, receiver, conv_type)
            await portal.insert()
            await portal.postinit()
            return portal

        return None

    async def get_dm_puppet(self) -> p.Puppet | None:
        if not self.is_direct:
            return None
        return await p.Puppet.get_by_twid(self.twid)

    # endregion



# Copyright (c) 2020 Tulir Asokan
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
from typing import Dict, Optional, Union

from attr import dataclass

from mautrix.types import SerializableAttrs, SerializableEnum
from mautrix.util import variation_selector

from .conversation import Conversation
from .util import StringTimestamp


class ReactionKey(SerializableEnum):
    FUNNY = "funny"  # 😂
    SURPRISED = "surprised"  # 😲
    SAD = "sad"  # 😢
    LIKE = "like"  # ❤
    EXCITED = "excited"  # 🔥
    AGREE = "agree"  # 👍
    DISAGREE = "disagree"  # 👎
    EMOJI = "emoji"  # arbitrary emoji

    @property
    def emoji(self) -> str:
        return _key_to_emoji[self]

    @classmethod
    def from_emoji(cls, emoji: str) -> Union["ReactionKey", str]:
        try:
            return _emoji_to_key[emoji.rstrip("\uFE0F")]
        except KeyError:
            return ReactionKey.EMOJI


_key_to_emoji: Dict[ReactionKey, str] = {
    ReactionKey.FUNNY: "\U0001F602",
    ReactionKey.SURPRISED: "\U0001F632",
    ReactionKey.SAD: "\U0001F622",
    ReactionKey.LIKE: "\u2764\uFE0F",
    ReactionKey.EXCITED: "\U0001F525",
    ReactionKey.AGREE: "\U0001F44D\uFE0F",
    ReactionKey.DISAGREE: "\U0001F44E\uFE0F",
}

_emoji_to_key: Dict[str, ReactionKey] = {
    "\U0001F44D": ReactionKey.AGREE,
    "\U0001F44E": ReactionKey.DISAGREE,
    "\u2764": ReactionKey.LIKE,
    "\U0001F525": ReactionKey.EXCITED,
    "\U0001F622": ReactionKey.SAD,
    "\U0001F632": ReactionKey.SURPRISED,
    "\U0001F62E": ReactionKey.SURPRISED,  # This is 😮, which is commonly used for surprised
    "\U0001F602": ReactionKey.FUNNY,
}


@dataclass
class ReactionCreateEntry(SerializableAttrs):
    id: str
    time: StringTimestamp
    conversation_id: str
    message_id: str
    reaction_key: ReactionKey
    sender_id: str
    emoji_reaction: Optional[str] = None
    affects_sort: Optional[bool] = None
    request_id: Optional[str] = None

    conversation: Optional[Conversation] = None

    @property
    def reaction_emoji(self) -> str:
        if self.reaction_key == ReactionKey.EMOJI:
            return self.emoji_reaction
        return self.reaction_key.emoji


@dataclass
class ReactionDeleteEntry(SerializableAttrs):
    id: str
    time: StringTimestamp
    conversation_id: str
    message_id: str
    reaction_key: ReactionKey
    emoji_reaction: str
    sender_id: str
    affects_sort: Optional[bool] = None

    conversation: Optional[Conversation] = None

    @property
    def reaction_emoji(self) -> str:
        if self.reaction_key == ReactionKey.EMOJI:
            emoji = self.emoji_reaction
        else:
            emoji = self.reaction_key.emoji
        return variation_selector.remove(emoji)



# Copyright (c) 2022 Tulir Asokan
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
from __future__ import annotations

from typing import Dict
import asyncio
import logging
import time

from aiohttp import ClientSession
from attr import dataclass
from yarl import URL

from . import conversation as c
from .dispatcher import TwitterDispatcher
from .errors import RateLimitError, TwitterAuthError, check_error
from .types import (
    Conversation,
    InboxTimeline,
    InitialStateResponse,
    PollResponse,
    TimelineStatus,
    User,
)


class PollingStarted:
    pass


class PollingStopped:
    pass


@dataclass
class PollingErrored:
    error: Exception
    fatal: bool
    count: int


class PollingErrorResolved:
    pass


class TwitterPoller(TwitterDispatcher):
    """This class handles polling for new messages using ``/dm/user_updates.json``."""

    dm_url: URL

    log: logging.Logger
    loop: asyncio.AbstractEventLoop
    http: ClientSession
    headers: dict[str, str]
    skip_poll_wait: asyncio.Event

    poll_sleep: int = 3
    error_sleep: int = 5
    max_poll_errors: int = 12
    poll_cursor: str | None
    dispatch_initial_resp: bool
    _poll_task: asyncio.Task | None
    _typing_in: c.Conversation | None

    @property
    def poll_params(self) -> dict[str, str]:
        return {
            "cards_platform": "Web-12",
            "include_cards": "1",
            "include_ext_alt_text": "true",
            "include_quote_count": "true",
            "include_reply_count": "1",
            "tweet_mode": "extended",
            "dm_users": "false",
            "include_groups": "true",
            "include_inbox_timelines": "true",
            "include_ext_media_color": "true",
            "supports_reactions": "true",
            "ext": ",".join(
                [
                    "mediaColor",
                    "altText",
                    "mediaStats",
                    "highlightedLabel",
                    "hasNftAvatar",
                    "voiceInfo",
                    "enrichments",
                    "superFollowMetadata",
                ]
            ),
        }

    @property
    def full_state_params(self) -> dict[str, str]:
        return {
            "include_profile_interstitial_type": "1",
            "include_blocking": "1",
            "include_blocked_by": "1",
            "include_followed_by": "1",
            "include_want_retweets": "1",
            "include_mute_edge": "1",
            "include_can_dm": "1",
            "include_can_media_tag": "1",
            "include_ext_alt_text": "1",
            "include_ext_has_nft_avatar": "1",
            "skip_status": "1",
            **self.poll_params,
        }

    async def all_trusted_conversations(
        self,
    ) -> tuple[Dict[str, Conversation], Dict[str, User]]:
        """
        Get all trusted conversations from the inbox (using pagination).

        Returns:
            Dictionary containing all conversations, keyed by their ID.
        """
        initial_state = await self.inbox_initial_state(set_poll_cursor=False)
        conversations = initial_state.conversations
        users = initial_state.users

        if initial_state.inbox_timelines.trusted.status == TimelineStatus.AT_END:
            return conversations, users

        min_entry_id = initial_state.inbox_timelines.trusted.min_entry_id

        while True:
            self.log.debug("Not at end, fetching more conversations with max_id %s", min_entry_id)
            inbox_timeline = await self.inbox_timeline("trusted", min_entry_id)

            if inbox_timeline.conversations is not None:
                conversations = {**conversations, **inbox_timeline.conversations}

            if inbox_timeline.users is not None:
                users = {**users, **inbox_timeline.users}

            if inbox_timeline.status == TimelineStatus.AT_END:
                return conversations, users

            if inbox_timeline.min_entry_id is not None:
                min_entry_id = inbox_timeline.min_entry_id
            else:
                return conversations, users

    async def inbox_timeline(self, inbox: str, max_id: str) -> InboxTimeline:
        """
        Gets a page of DMs from one of the possible inbox timelines.

        Returns:
            The response data from the server.
        """
        url = (self.dm_url / "inbox_timeline" / f"{inbox}.json").with_query(
            {**self.full_state_params, "max_id": max_id}
        )
        async with self.http.get(url, headers=self.headers) as resp:
            data = await check_error(resp)
            response = InboxTimeline.deserialize(data["inbox_timeline"])
            return response

    async def inbox_initial_state(self, set_poll_cursor: bool = True) -> InitialStateResponse:
        """
        Get the initial DM inbox state, including conversations, user profiles and some messages.

        This also gets the initial :attr:`poll_cursor` value.

        Returns:
            The response data from the server.
        """
        url = (self.dm_url / "inbox_initial_state.json").with_query(
            {
                **self.full_state_params,
                "filter_low_quality": "true",
                "include_quality": "all",
            }
        )
        async with self.http.get(url, headers=self.headers) as resp:
            data = await check_error(resp)
            response = InitialStateResponse.deserialize(data["inbox_initial_state"])
            if set_poll_cursor:
                self.poll_cursor = response.cursor
            return response

    async def _poll_once(self) -> PollResponse:
        if not self.poll_cursor:
            raise RuntimeError("Cursor must be set to poll")
        url = (self.dm_url / "user_updates.json").with_query(
            {
                **self.poll_params,
                "cursor": self.poll_cursor,
                "filter_low_quality": "true",
                "include_quality": "all",
            }
        )
        async with self.http.get(url, headers=self.headers) as resp:
            data = await check_error(resp)
            try:
                user_events = data["user_events"]
            except KeyError:
                try:
                    inbox_initial_state = data["inbox_initial_state"]
                except KeyError:
                    self.log.warning("Got unknown poll response: %s", data)
                    raise
                else:
                    response = InitialStateResponse.deserialize(inbox_initial_state)
            else:
                response = PollResponse.deserialize(user_events)
            self.poll_cursor = response.cursor
            return response

    async def poll_forever(self, raise_exceptions: bool = True) -> None:
        """
        Poll for updates forever.

        Args:
            raise_exceptions: Whether or not fatal errors should be raised after logging.
                :class:`asyncio.CancelledError` will not be raised in any case.
        """
        try:
            await self._poll_forever()
        except asyncio.CancelledError:
            self.log.debug("Polling stopped")
            await self.dispatch(PollingStopped())
        except TwitterAuthError as e:
            self.log.exception("Auth error while polling")
            await self.dispatch(PollingErrored(e, fatal=True, count=0))
        except Exception as e:
            await self.dispatch(PollingErrored(e, fatal=True, count=0))
            self.log.exception("Fatal error while polling")
            if raise_exceptions:
                raise

    async def dispatch_all(self, resp: PollResponse | InitialStateResponse) -> None:
        for user in (resp.users or {}).values():
            await self.dispatch(user)
        for conversation in (resp.conversations or {}).values():
            await self.dispatch(conversation)
        for entry in resp.entries or []:
            if not entry:
                continue
            for entry_type in entry.all_types:
                try:
                    entry_type.conversation = resp.conversations[entry_type.conversation_id]
                except KeyError:
                    msg = (
                        "Poll response didn't contain conversation info "
                        f"for {entry_type.conversation_id} "
                        f"(entry type {type(entry_type)}, ID {entry_type.id}"
                    )
                    if entry_type.conversation:
                        self.log.debug(f"{msg}, but the entry had its own conversation info")
                        await self.dispatch(entry_type)
                    else:
                        # TODO we could still dispatch the event and guess the conversation type
                        #      in event handlers if necessary.
                        self.log.warning(msg)
                else:
                    await self.dispatch(entry_type)

    async def _poll_forever(self) -> None:
        if not self.poll_cursor:
            self.log.debug("Poll cursor not set, calling initial state to get cursor")
            resp = await self.inbox_initial_state()
            if self.dispatch_initial_resp:
                await self.dispatch_all(resp)
        await self.dispatch(PollingStarted())
        errors = 0
        while True:
            try:
                resp = await self._poll_once()
            except RateLimitError as e:
                sleep = e.reset - int(time.time())
                self.log.warning(
                    f"Got rate limit until {e.reset} while polling, waiting {sleep} seconds"
                )
                if self.poll_sleep < 8:
                    self.poll_sleep += 1
                    self.log.debug(f"Increased poll sleep to {self.poll_sleep}")
                await asyncio.sleep(sleep)
                continue
            except TwitterAuthError:
                raise
            except Exception as e:
                if errors > self.max_poll_errors > 0:
                    self.log.debug(
                        f"Error count ({errors}) exceeded maximum, raising error as fatal"
                    )
                    raise
                errors += 1
                sleep = min(15 * 60, self.error_sleep * errors)
                self.log.warning(f"Error while polling, retrying in {sleep}s", exc_info=True)
                await self.dispatch(PollingErrored(e, fatal=False, count=errors))
                await asyncio.sleep(sleep)
                continue
            if errors > 0:
                errors = 0
                await self.dispatch(PollingErrorResolved())
            await self.dispatch_all(resp)
            try:
                await asyncio.wait_for(self.skip_poll_wait.wait(), self.poll_sleep)
            except asyncio.TimeoutError:
                pass
            if self._typing_in:
                await self._typing_in.mark_typing()

    def is_polling(self) -> bool:
        return self._poll_task and not self._poll_task.done()

    def start_polling(self) -> asyncio.Task:
        """
        Start polling forever in the background. This calls :meth:`poll_forever` and puts it in an
        asyncio Task. The task is stored so it can be cancelled with :meth:`stop_polling`.

        Returns:
            The created asyncio task.
        """
        self.stop_polling()
        self.log.debug("Starting poll task")
        self._poll_task = asyncio.create_task(self.poll_forever())
        return self._poll_task

    def stop_polling(self) -> None:
        """Stop the ongoing poll task. Any ongoing handlers will also be cancelled."""
        if self._poll_task:
            self.log.debug("Cancelling ongoing poll task")
            self._poll_task.cancel()
            self._poll_task = None



# mautrix-twitter - A Matrix-Twitter DM puppeting bridge
# Copyright (C) 2022 Tulir Asokan
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
from __future__ import annotations

from typing import TYPE_CHECKING, AsyncGenerator, AsyncIterable, Awaitable, cast
import asyncio
import logging

from mautrix.appservice import DOUBLE_PUPPET_SOURCE_KEY
from mautrix.bridge import BaseUser, async_getter_lock
from mautrix.errors import MNotFound
from mautrix.types import (
    EventID,
    MessageType,
    PushActionType,
    PushRuleKind,
    PushRuleScope,
    RoomID,
    RoomTagInfo,
    TextMessageEventContent,
    UserID,
)
from mautrix.util import background_task
from mautrix.util.bridge_state import BridgeState, BridgeStateEvent
from mautrix.util.opt_prometheus import Gauge, Summary, async_time
from mautwitdm import TwitterAPI
from mautwitdm.errors import TwitterAuthError, TwitterError
from mautwitdm.poller import PollingErrored, PollingErrorResolved, PollingStarted, PollingStopped
from mautwitdm.types import (
    Conversation,
    ConversationReadEntry,
    MessageEntry,
    ReactionCreateEntry,
    ReactionDeleteEntry,
    User as TwitterUser,
)

from . import portal as po, puppet as pu
from .config import Config
from .db import Portal as DBPortal, User as DBUser

if TYPE_CHECKING:
    from .__main__ import TwitterBridge

METRIC_CONVERSATION_UPDATE = Summary(
    "bridge_on_conversation_update", "calls to handle_conversation_update"
)
METRIC_USER_UPDATE = Summary("bridge_on_user_update", "calls to handle_user_update")
METRIC_MESSAGE = Summary("bridge_on_message", "calls to handle_message")
METRIC_REACTION = Summary("bridge_on_reaction", "calls to handle_reaction")
METRIC_RECEIPT = Summary("bridge_on_receipt", "calls to handle_receipt")
METRIC_LOGGED_IN = Gauge("bridge_logged_in", "Users logged into the bridge")
METRIC_CONNECTED = Gauge("bridge_connected", "Bridged users connected to Twitter")

BridgeState.human_readable_errors.update(
    {
        "logged-out": "You're not logged into Twitter",
        "twitter-connection-failed": "Failed to connect to Twitter",
        "twitter-not-connected": None,
        "twitter-connection-error": "An error occurred while polling Twitter for new messages",
    }
)


class User(DBUser, BaseUser):
    by_mxid: dict[UserID, User] = {}
    by_twid: dict[int, User] = {}
    config: Config

    client: TwitterAPI | None

    permission_level: str
    username: str | None

    _notice_room_lock: asyncio.Lock
    _notice_send_lock: asyncio.Lock
    _is_logged_in: bool | None
    _connected: bool
    _intentional_stop: bool
    _connect_task: asyncio.Task | None

    def __init__(
        self,
        mxid: UserID,
        twid: int | None = None,
        auth_token: str | None = None,
        csrf_token: str | None = None,
        poll_cursor: str | None = None,
        notice_room: RoomID | None = None,
    ) -> None:
        super().__init__(
            mxid=mxid,
            twid=twid,
            auth_token=auth_token,
            csrf_token=csrf_token,
            poll_cursor=poll_cursor,
            notice_room=notice_room,
        )
        BaseUser.__init__(self)
        self._notice_room_lock = asyncio.Lock()
        self._notice_send_lock = asyncio.Lock()
        perms = self.config.get_permissions(mxid)
        self.is_whitelisted, self.is_admin, self.permission_level = perms
        self.client = None
        self.username = None
        self._is_logged_in = None
        self._connected = False
        self._intentional_stop = False
        self._connect_task = None

    @classmethod
    def init_cls(cls, bridge: "TwitterBridge") -> AsyncIterable[Awaitable[None]]:
        cls.bridge = bridge
        cls.config = bridge.config
        cls.az = bridge.az
        cls.loop = bridge.loop
        TwitterAPI.error_sleep = cls.config["bridge.error_sleep"]
        TwitterAPI.max_poll_errors = cls.config["bridge.max_poll_errors"]
        return (user.try_connect() async for user in cls.all_logged_in())

    async def update(self) -> None:
        if self.client:
            self.auth_token, self.csrf_token = self.client.tokens
            self.poll_cursor = self.client.poll_cursor
        await super().update()

    # region Connection management

    async def is_logged_in(self, ignore_cache: bool = False) -> bool:
        if not self.client:
            return False
        if self._is_logged_in is None:
            try:
                self._is_logged_in = await self.client.get_user_identifier() is not None
            except Exception:
                self._is_logged_in = False
        return self.client and self._is_logged_in

    async def get_puppet(self) -> pu.Puppet | None:
        if not self.twid:
            return None
        return await pu.Puppet.get_by_twid(self.twid)

    async def try_connect(self) -> None:
        try:
            await self._connect()
        except TwitterAuthError as e:
            self.log.exception("Auth error while connecting to Twitter")
            await self.push_bridge_state(
                BridgeStateEvent.BAD_CREDENTIALS,
                error="twitter-auth-error",
                message=e.message,
            )
        except TwitterError as e:
            self.log.exception("Error while connecting to Twitter")
            await self.push_bridge_state(
                BridgeStateEvent.UNKNOWN_ERROR,
                error="twitter-unknown-error",
                message=e.message,
            )
        except Exception:
            self.log.exception("Unknown exception while connecting to Twitter")
            await self.push_bridge_state(
                BridgeStateEvent.UNKNOWN_ERROR, error="twitter-connection-failed"
            )

    async def locked_connect(self, auth_token: str, csrf_token: str) -> None:
        if self._connect_task and not self._connect_task.done():
            self.log.warning("locked_connect() called when another call was already running")
            await self._connect_task
        elif self.auth_token == auth_token and self.csrf_token == csrf_token:
            self.log.warning("locked_connect() called with the credentials already in use")
        else:
            self._connect_task = asyncio.create_task(self._connect(auth_token, csrf_token))
            try:
                await self._connect_task
            finally:
                self._connect_task = None

    async def _connect(self, auth_token: str | None = None, csrf_token: str | None = None) -> None:
        client = TwitterAPI(
            log=logging.getLogger("mau.twitter.api").getChild(self.mxid),
            loop=self.loop,
            node_id=hash(self.mxid) % (2**48),
        )
        client.poll_cursor = self.poll_cursor
        client.set_tokens(auth_token or self.auth_token, csrf_token or self.csrf_token)

        # Initial ping to make sure auth works
        await client.get_user_identifier()

        self.client = client
        self.client.add_handler(Conversation, self.handle_conversation_update)
        self.client.add_handler(TwitterUser, self.handle_user_update)
        self.client.add_handler(MessageEntry, self.handle_message)
        self.client.add_handler(ReactionCreateEntry, self.handle_reaction)
        self.client.add_handler(ReactionDeleteEntry, self.handle_reaction)
        self.client.add_handler(ConversationReadEntry, self.handle_receipt)
        self.client.add_handler(PollingStarted, self.on_connect)
        self.client.add_handler(PollingErrorResolved, self.on_connect)
        self.client.add_handler(PollingStopped, self.on_disconnect)
        self.client.add_handler(PollingErrored, self.on_disconnect)
        self.client.add_handler(PollingErrored, self.on_error)
        self.client.add_handler(PollingErrorResolved, self.on_error_resolved)

        user_info = await self.get_info()
        self.twid = user_info.id
        self._track_metric(METRIC_LOGGED_IN, True)
        self.by_twid[self.twid] = self

        await self.update()

        self._intentional_stop = False
        if self.poll_cursor:
            self.log.debug("Poll cursor set, starting polling right away (not initial syncing)")
            self.client.start_polling()
        else:
            background_task.create(self._try_initial_sync())
        background_task.create(self._try_sync_puppet(user_info))

    async def fill_bridge_state(self, state: BridgeState) -> None:
        await super().fill_bridge_state(state)
        if self.twid:
            state.remote_id = str(self.twid)
            puppet = await pu.Puppet.get_by_twid(self.twid)
            state.remote_name = puppet.name

    async def get_bridge_states(self) -> list[BridgeState]:
        if not self.twid:
            return []
        state = BridgeState(state_event=BridgeStateEvent.UNKNOWN_ERROR)
        if self._connected:
            state.state_event = BridgeStateEvent.CONNECTED
        return [state]

    async def on_connect(self, evt: PollingStarted | PollingErrorResolved) -> None:
        self._track_metric(METRIC_CONNECTED, True)
        self._connected = True
        await self.push_bridge_state(BridgeStateEvent.CONNECTED)

    async def on_disconnect(self, evt: PollingStopped | PollingErrored) -> None:
        self._track_metric(METRIC_CONNECTED, False)
        self._connected = False
        if isinstance(evt, PollingStopped) and not self._intentional_stop:
            await self.push_bridge_state(
                BridgeStateEvent.UNKNOWN_ERROR, error="twitter-not-connected"
            )
        self._intentional_stop = False

    # TODO this stuff could probably be moved to mautrix-python
    async def get_notice_room(self) -> RoomID:
        if not self.notice_room:
            async with self._notice_room_lock:
                # If someone already created the room while this call was waiting,
                # don't make a new room
                if self.notice_room:
                    return self.notice_room
                creation_content = {}
                if not self.config["bridge.federate_rooms"]:
                    creation_content["m.federate"] = False
                self.notice_room = await self.az.intent.create_room(
                    is_direct=True,
                    invitees=[self.mxid],
                    topic="Twitter DM bridge notices",
                    creation_content=creation_content,
                )
                await self.update()
        return self.notice_room

    async def send_bridge_notice(
        self,
        text: str,
        edit: EventID | None = None,
        state_event: BridgeStateEvent | None = None,
        important: bool = False,
    ) -> EventID | None:
        if state_event:
            await self.push_bridge_state(state_event, message=text)
        if self.config["bridge.disable_bridge_notices"]:
            return None
        event_id = None
        try:
            self.log.debug("Sending bridge notice: %s", text)
            content = TextMessageEventContent(
                body=text,
                msgtype=(MessageType.TEXT if important else MessageType.NOTICE),
            )
            if edit:
                content.set_edit(edit)
            # This is locked to prevent notices going out in the wrong order
            async with self._notice_send_lock:
                event_id = await self.az.intent.send_message(await self.get_notice_room(), content)
        except Exception:
            self.log.warning("Failed to send bridge notice", exc_info=True)
        return edit or event_id

    async def on_error(self, evt: PollingErrored) -> None:
        if evt.fatal:
            if isinstance(evt.error, TwitterAuthError):
                await self.push_bridge_state(
                    BridgeStateEvent.BAD_CREDENTIALS,
                    error="twitter-auth-error",
                    message=evt.error.message,
                )
                await self.send_bridge_notice(
                    f"Auth error while polling Twitter: {evt.error}", important=True
                )
            else:
                await self.send_bridge_notice(
                    f"Fatal error while polling Twitter: {evt.error}",
                    state_event=BridgeStateEvent.UNKNOWN_ERROR,
                    important=evt.fatal,
                )
        elif evt.count == 1 and self.config["bridge.temporary_disconnect_notices"]:
            await self.send_bridge_notice(
                f"Error while polling Twitter: {evt.error}\nThe bridge will keep retrying.",
                state_event=BridgeStateEvent.TRANSIENT_DISCONNECT,
            )
        else:
            state_event = (
                BridgeStateEvent.TRANSIENT_DISCONNECT
                if evt.count < 5
                else BridgeStateEvent.UNKNOWN_ERROR
            )
            await self.push_bridge_state(state_event, f"Error while polling Twitter: {evt.error}")

    async def on_error_resolved(self, evt: PollingErrorResolved) -> None:
        if self.config["bridge.temporary_disconnect_notices"]:
            await self.send_bridge_notice(
                "Twitter polling error resolved", state_event=BridgeStateEvent.CONNECTED
            )

    async def _try_sync_puppet(self, user_info: TwitterUser) -> None:
        puppet = await pu.Puppet.get_by_twid(self.twid)
        try:
            await puppet.update_info(user_info)
        except Exception:
            self.log.exception("Failed to update own puppet info")
        try:
            if puppet.custom_mxid != self.mxid and puppet.can_auto_login(self.mxid):
                self.log.info(f"Automatically enabling custom puppet")
                await puppet.switch_mxid(access_token="auto", mxid=self.mxid)
        except Exception:
            self.log.exception("Failed to automatically enable custom puppet")

    async def _try_initial_sync(self) -> None:
        try:
            await self.sync()
        except Exception:
            self.log.exception("Exception while syncing conversations")
        self.log.debug("Initial sync completed, starting polling")
        self.client.start_polling()

    async def get_direct_chats(self) -> dict[UserID, list[RoomID]]:
        return {
            pu.Puppet.get_mxid_from_id(portal.other_user): [portal.mxid]
            for portal in await DBPortal.find_private_chats_of(self.twid)
            if portal.mxid
        }

    async def get_portal_with(self, puppet: pu.Puppet, create: bool = True) -> po.Portal | None:
        # We should probably make this work eventually, but for now, creating chats will just not
        # work.
        return None

    async def sync(self) -> None:
        await self.push_bridge_state(BridgeStateEvent.BACKFILLING)
        resp = await self.client.inbox_initial_state(set_poll_cursor=False)
        if not self.poll_cursor:
            self.poll_cursor = resp.cursor
        self.client.poll_cursor = self.poll_cursor
        self.log.debug("Fetching all trusted conversations...")
        conversations, users = await self.client.all_trusted_conversations()
        limit = self.config["bridge.initial_conversation_sync"]
        conversations = sorted(
            conversations.values(), key=lambda conv: conv.sort_timestamp, reverse=True
        )
        self.log.info("Got %d conversations (sync limit %d)", len(conversations), limit)
        if limit < 0:
            limit = len(conversations)
        for _, user in users.items():
            await self.handle_user_update(user)
        index = 0
        for conversation in conversations:
            self.log.info(
                "Syncing conversation %s (%d of %d)", conversation.conversation_id, index, limit
            )
            try:
                create_portal = index < limit and conversation.trusted
                if create_portal:
                    index += 1
                await self.handle_conversation_update(conversation, create_portal=create_portal)
            except Exception:
                self.log.exception(
                    "Error while syncing conversation %s!", conversation.conversation_id
                )
        await self.update_direct_chats()

    async def tag_room(self, puppet: pu.Puppet, portal: po.Portal, tag: str, active: bool) -> None:
        if not tag or not portal or not portal.mxid:
            return
        tag_info = await puppet.intent.get_room_tag(portal.mxid, tag)
        if active and tag_info is None:
            tag_info = RoomTagInfo(order=0.5)
            tag_info[DOUBLE_PUPPET_SOURCE_KEY] = self.bridge.name
            self.log.debug(f"Adding tag {tag} to {portal.mxid}/{portal.twid}")
            await puppet.intent.set_room_tag(portal.mxid, tag, tag_info)
        elif (
            not active and tag_info and tag_info.get(DOUBLE_PUPPET_SOURCE_KEY) == self.bridge.name
        ):
            self.log.debug(f"Removing tag {tag} from {portal.mxid}/{portal.twid}")
            await puppet.intent.remove_room_tag(portal.mxid, tag)

    async def set_muted(self, puppet: pu.Puppet, portal: po.Portal, muted: bool) -> None:
        if not portal or not portal.mxid:
            return
        if muted:
            self.log.debug(f"Muting {portal.mxid}/{portal.twid}")
            await puppet.intent.set_push_rule(
                PushRuleScope.GLOBAL,
                PushRuleKind.ROOM,
                portal.mxid,
                actions=[PushActionType.DONT_NOTIFY],
            )
        else:
            try:
                await puppet.intent.remove_push_rule(
                    PushRuleScope.GLOBAL, PushRuleKind.ROOM, portal.mxid
                )
                self.log.debug(f"Unmuted {portal.mxid}/{portal.twid}")
            except MNotFound:
                pass

    async def get_info(self) -> TwitterUser:
        settings = await self.client.get_settings()
        self.username = settings["screen_name"]
        return (await self.client.lookup_users(usernames=[self.username]))[0]

    async def stop(self) -> None:
        if self.client:
            self._intentional_stop = True
            self.client.stop_polling()
        self._track_metric(METRIC_CONNECTED, False)
        await self.update()

    async def logout(self) -> None:
        if self.client:
            self._intentional_stop = True
            self.client.stop_polling()
        self._track_metric(METRIC_CONNECTED, False)
        self._track_metric(METRIC_LOGGED_IN, False)
        puppet = await pu.Puppet.get_by_twid(self.twid, create=False)
        if puppet and puppet.is_real_user:
            await puppet.switch_mxid(None, None)
        try:
            del self.by_twid[self.twid]
        except KeyError:
            pass
        self.client = None
        self._is_logged_in = None
        self.twid = None
        self.poll_cursor = None
        self.auth_token = None
        self.csrf_token = None
        await self.update()

    # endregion
    # region Event handlers

    @async_time(METRIC_CONVERSATION_UPDATE)
    async def handle_conversation_update(
        self, evt: Conversation, create_portal: bool = False
    ) -> None:
        portal = await po.Portal.get_by_twid(
            evt.conversation_id, receiver=self.twid, conv_type=evt.type
        )
        if not portal.mxid:
            self.log.debug("Conversation %s doesn't have MXID!", evt.conversation_id)
            if create_portal:
                self.log.debug("Creating Matrix room...")
                await portal.create_matrix_room(self, evt)
        else:
            # We don't want to do the invite_user and such things each time conversation info
            # comes down polling, so if the room already exists, only call .update_info()
            await portal.update_info(evt)
            puppet = await pu.Puppet.get_by_custom_mxid(self.mxid)
            if puppet:
                if self.config["bridge.low_quality_tag"]:
                    self.log.debug("Tagging room if low-quality")
                    await self.tag_room(
                        puppet,
                        portal,
                        self.config["bridge.low_quality_tag"],
                        evt.low_quality == True,
                    )
                if self.config["bridge.low_quality_mute"] and evt.low_quality:
                    await self.set_muted(puppet, portal, True)

    @async_time(METRIC_USER_UPDATE)
    async def handle_user_update(self, user: TwitterUser) -> None:
        puppet = await pu.Puppet.get_by_twid(user.id)
        await puppet.update_info(user)

    @async_time(METRIC_MESSAGE)
    async def handle_message(self, evt: MessageEntry) -> None:
        portal = await po.Portal.get_by_twid(
            evt.conversation_id, receiver=self.twid, conv_type=evt.conversation.type
        )
        if not portal.mxid:
            await portal.create_matrix_room(self, evt.conversation)
        sender = await pu.Puppet.get_by_twid(int(evt.message_data.sender_id))
        await portal.backfill_lock.wait(f"{evt.message_data.sender_id}/{evt.message_data.id}")
        await portal.handle_twitter_message(self, sender, evt.message_data, evt.request_id)

    @async_time(METRIC_REACTION)
    async def handle_reaction(self, evt: ReactionCreateEntry | ReactionDeleteEntry) -> None:
        portal = await po.Portal.get_by_twid(
            evt.conversation_id, receiver=self.twid, conv_type=evt.conversation.type
        )
        if not portal.mxid:
            self.log.debug(f"Ignoring reaction in conversation {evt.conversation_id} with no room")
            return
        puppet = await pu.Puppet.get_by_twid(int(evt.sender_id))
        if isinstance(evt, ReactionCreateEntry):
            await portal.handle_twitter_reaction_add(
                puppet, int(evt.message_id), evt.reaction_emoji, evt.time, int(evt.id)
            )
        else:
            await portal.handle_twitter_reaction_remove(
                puppet, int(evt.message_id), evt.reaction_emoji
            )

    @async_time(METRIC_RECEIPT)
    async def handle_receipt(self, evt: ConversationReadEntry) -> None:
        portal = await po.Portal.get_by_twid(
            evt.conversation_id, receiver=self.twid, conv_type=evt.conversation.type
        )
        if not portal.mxid:
            return
        sender = await pu.Puppet.get_by_twid(self.twid)
        await portal.handle_twitter_receipt(sender, int(evt.last_read_event_id), historical=False)

    # endregion
    # region Database getters

    def _add_to_cache(self) -> None:
        self.by_mxid[self.mxid] = self
        if self.twid:
            self.by_twid[self.twid] = self

    @classmethod
    @async_getter_lock
    async def get_by_mxid(cls, mxid: UserID, *, create: bool = True) -> User | None:
        # Never allow ghosts to be users
        if pu.Puppet.get_id_from_mxid(mxid):
            return None
        try:
            return cls.by_mxid[mxid]
        except KeyError:
            pass

        user = cast(cls, await super().get_by_mxid(mxid))
        if user is not None:
            user._add_to_cache()
            return user

        if create:
            user = cls(mxid)
            await user.insert()
            user._add_to_cache()
            return user

        return None

    @classmethod
    @async_getter_lock
    async def get_by_twid(cls, twid: int) -> User | None:
        try:
            return cls.by_twid[twid]
        except KeyError:
            pass

        user = cast(cls, await super().get_by_twid(twid))
        if user is not None:
            user._add_to_cache()
            return user

        return None

    @classmethod
    async def all_logged_in(cls) -> AsyncGenerator[User, None]:
        users = await super().all_logged_in()
        user: cls
        for index, user in enumerate(users):
            try:
                yield cls.by_mxid[user.mxid]
            except KeyError:
                user._add_to_cache()
                yield user

    # endregion



from .conversation import Conversation, ConversationType, Participant
from .entry import (
    ConversationCreateEntry,
    ConversationNameUpdateEntry,
    ConversationReadEntry,
    Entry,
    TrustConversationEntry,
)
from .message import MessageData, MessageEntry
from .message_attachment import (
    RGB,
    CardBindingValue,
    ExtendedMessageEntities,
    ImageBindingValue,
    ImageColor,
    ImageColorBindingValue,
    MessageAttachment,
    MessageAttachmentCard,
    MessageAttachmentMedia,
    MessageAttachmentTweet,
    MessageEntityMedia,
    TweetAttachmentStatus,
    VideoInfo,
    VideoVariant,
)
from .message_entity import (
    MessageEntities,
    MessageEntitySimple,
    MessageEntityURL,
    MessageEntityUserMention,
)
from .reaction import ReactionCreateEntry, ReactionDeleteEntry, ReactionKey
from .response import (
    FetchConversationResponse,
    InboxTimeline,
    InitialStateResponse,
    MediaUploadResponse,
    PollResponse,
    SendResponse,
    TimelineStatus,
)
from .stream_payload import DMTypingEvent, DMUpdateEvent, StreamEvent
from .user import User
